"""
This class is to make complex testings.
"""
from yta_multimedia.video.generation.manim.classes.base_manim_animation import BaseManimAnimation
from yta_multimedia.video.generation.manim.enums import MANIM_RENDERER
from yta_multimedia.video.generation.manim.classes.base_manim_animation_wrapper import BaseManimAnimationWrapper, ManimAnimationType
from manim import *
from typing import Union


class TestTextManimAnimationWrapper(BaseManimAnimationWrapper):
    """
    Simple Text animation wrapper, used only for testing.
    """
    text: str = None
    duration: float = None

    def __init__(self, text: str, duration: float):
        # TODO: This parameter validation could be done
        # by our specific parameter validator (under
        # construction yet)
        exception_messages = []

        if len(exception_messages) > 0:
            raise Exception('\n'.join([exception_message for exception_message in exception_messages]))
        
        self.text = text
        self.duration = duration
        super().__init__(TestTextManimAnimationGenerator, ManimAnimationType.TEXT_ALPHA)

class TestTextManimAnimationGenerator(BaseManimAnimation):
    def construct(self):
        """
        This method is called by manim when executed by shell and
        will call the scene animation render method to be processed
        and generated.
        """
        self.animate()

    def generate(self, parameters: dict, output_filename: Union[str, None] = None):
        """
        Build the manim animation video and stores it
        locally as the provided 'output_filename', 
        returning the final output filename.

        This method will call the '.animate()' method
        to build the scene with the instructions that
        are written in that method.
        """
        return super().generate(
            parameters,
            renderer = MANIM_RENDERER.CAIRO,
            output_filename = output_filename
        )
    
    def animate(self):
        """
        This code will generate the manim animation and belongs to the
        Scene manim object.
        """
        text = Text('Texto de prueba')

        self.add(text)
        self.wait(1)

        # TODO: Move this to a utils: this make it dissapear like an helicopter
        # def x(mob, alpha):
        #     mob.shift(interpolate(0, 1, alpha)).rotate(interpolate(0, 90 * DEGREES, alpha))
        # self.play(
        #     UpdateFromAlphaFunc(text, x)
        # )


        # def x(mob, alpha):
        #     mob.shift(interpolate(0, 1 * UP, alpha)).rotate(interpolate(0, 90 * DEGREES, alpha))
        # self.play(UpdateFromAlphaFunc(text, x), run_time = 2)

        #self.play(text.animate.scale(2).rotate(90 * DEGREES), run_time = 2, rate_func = linear)

        # TODO: This plays the two animaions
        # self.play(
        #     Succession(
        #         ApplyMethod(text.scale, 2),
        #         ApplyMethod(text.scale, 0.5),
        #     ), # this is equivalent to the one below (please check manim code)
        #     text.animate.rotate(90 * DEGREES),
        #     run_time = 2,
        #     rate_func = linear
        # )

        # # TODO: This plays only the last animation
        # # Simple form
        # self.play(
        #     text.animate.scale(2), 
        #     text.animate.rotate(-90 * DEGREES), # Even if they are different
        #     run_time = 1
        # )

        # This plays the two animations at the same time
        # self.play(
        #     text.animate.scale(2).rotate(-90 * DEGREES),
        #     run_time = 1
        # )

        # This saves a single animation (with specific runtime) to be played later
        # anim = text.animate.shift(RIGHT * 2).set_run_time(2)
        # self.play(
        #     anim,
        #     run_time = 1
        # )

        # TODO: This only plays the last animation
        # self.play(
        #     Succession(
        #         text.animate.rotate(90 * DEGREES).set_runtime(1)
        #         text.animate.rotate(90 * DEGREES).set_runtime(2)
        #         text.animate.rotate(90 * DEGREES).set_runtime(1)
        #         text.animate.rotate(90 * DEGREES).set_runtime(2) # This
        #     ),
        #     run_time = 1
        # )

        def play_animations(animations, text):
            """
            Returns all the animations that must be played for each frame
            (with 'run_time = 1 / 60'). If an animation is None, a wait 
            must be played.
            """
            # I Have animations that include the name, the params, the moments
            # and I have to make animations frame by frame to be able to handle
            # Example: rotate 360 from 0 to 1.5s (frames 0 to 90), shift 3 from 0.75
            # to 1.24s (frames 75 to 105)
            animations = [
                {
                    'mobject': text,
                    'name': 'rotate',
                    'params': 360 * DEGREES,
                    'time': [0, 1.5]
                },
                {
                    'mobject': text,
                    'name': 'shift',
                    'params': 3 * LEFT,
                    'time': [0.75, 1.25]
                }
            ]

            animations_to_play = []
            # TODO: Obtain the greates 'time'
            greatest_time = 1.5 # Manually set by now
            FPS = 60
            for frame in range(int(FPS * greatest_time)):
                animation_to_add = None
                for animation in animations:
                    if frame >= (animation['time'][0] * FPS) and frame < (animation['time'][1] * FPS):
                        # TODO: What about non-divisible animations or progressive ones (?)
                        # It is not the same to play a scale 2x once that play consecutive
                        # scalations...
                        params = animation['params']
                        # Param must be according to the number of frames
                        # We substract 1 frame because if you have 3 frames you can only make
                        # 2 animations: the 0-1 and the 1-2, so if you want to rotate
                        # from 0 to 360 in 3 frames, yo cannot do 120-120-120, you only
                        # have 2 animations, so it must be 180-180. Thats the reason of
                        # the - (1 / 60)
                        params = params / ((animation['time'][1] - animation['time'][0] - (1 / 60)) * 60)
                        print(params)
                        if animation_to_add is None:
                            # Dynamically call the animation method
                            # print(animation['mobject'].animate)
                            # func = getattr(animation['mobject'].animate, animation['name'])
                            # print(func)
                            # method_name = "rotate"  # Por ejemplo, 'rotate' o 'shift'
                            # method = getattr(animation['mobject'].animate, method_name)
                            # print(method)
                            #self.play(method(*method_args))
                            animation_to_add = getattr(animation['mobject'].animate, animation['name'])(params)
                        else:
                            animation_to_add = getattr(animation_to_add, animation['name'])(params)
                    
                print(f'frame {str(frame)}')
                if animation_to_add is not None:
                    self.play(animation_to_add, run_time = 1 / 60, rate_func = linear)
                    #animations_to_play.append(animation_to_add)
                else:
                    self.wait(1 / 60)
                    #animations_to_play.append(None)

            #return animations_to_play

        # This is working, slow, but working. Maybe it could be a way of
        # handle a timeline with animations so we just define when and how
        # and they are built with this automatic system.
        #animations_to_play = play_animations('', text)

        kwargs = {
            'color': BLUE,
            'font_size': 22
        }
        self.play(Text('hola', **kwargs).animate.rotate(90 * DEGREES), run_time = 2)

        # This doesn't work as expected
        # self.play(text.animate.rotate(360 * DEGREES * (2 / 4)), run_time = 2)
        # self.play(text.animate.rotate(360 * DEGREES * (1 / 4)).shift(2 * LEFT), run_time = 1)
        # self.play(text.animate.rotate(360 * DEGREES * (1 / 4)), run_time = 1)

        # for animation in animations_to_play:
        #     if animation is None:
        #         self.wait(1 / 60)
        #     else:
        #         print(animation)
        #         self.play(animation, run_time = 1 / 60)

        # TODO: This below is working, but not if stored in array with method
        # for frame in range(int(60 * 0.5)):
        #     x = getattr(text.animate, 'rotate')(45 * DEGREES / (60 * 0.5))
        #     #x = text.animate.rotate(45 * DEGREES / (60 * 0.5))
        #     if frame > 5 and frame < 15:
        #         x = getattr(x, 'shift')(0.5 * LEFT / (60 * 0.5))
        #         #x = x.shift(0.5 * LEFT / (60 * 0.5))

        #     self.play(x, run_time = 1 / 60)
        #     #self.play(text.animate.rotate(45 * DEGREES / (60 * 0.5)).shift(0.5 * LEFT / (60 * 0.5)), run_time = 1 / 60)

        # self.play(text.animate.rotate(45 * DEGREES).shift(0.5 * LEFT), run_time = 1 / 8)
        # self.play(text.animate.rotate(45 * DEGREES).shift(0.5 * LEFT), run_time = 1 / 8)
        # self.play(text.animate.rotate(45 * DEGREES).shift(0.5 * LEFT), run_time = 1 / 8)
        # self.play(text.animate.rotate(45 * DEGREES).shift(0.5 * LEFT), run_time = 1 / 8)
        # self.play(text.animate.rotate(45 * DEGREES).shift(0.5 * LEFT), run_time = 1 / 8)
        # self.play(text.animate.rotate(45 * DEGREES).shift(0.5 * LEFT), run_time = 1 / 8)
        # self.play(text.animate.rotate(45 * DEGREES).shift(0.5 * LEFT), run_time = 1 / 8)
        # self.play(text.animate.rotate(45 * DEGREES).shift(0.5 * LEFT), run_time = 1 / 8)

        # self.play(
        #     Succession(
        #         ApplyMethod(text.rotate, 45 * DEGREES, run_time = 1),
        #         ApplyMethod(text.rotate, 45 * DEGREES, run_time = 1),
        #         ApplyMethod(text.rotate, 45 * DEGREES, run_time = 1),
        #         ApplyMethod(text.rotate, 45 * DEGREES, run_time = 1),
        #         ApplyMethod(text.rotate, 45 * DEGREES, run_time = 1),
        #         ApplyMethod(text.rotate, 45 * DEGREES, run_time = 1),
        #         ApplyMethod(text.rotate, 45 * DEGREES, run_time = 1),
        #         ApplyMethod(text.rotate, 45 * DEGREES, run_time = 1),
        #     ),
        #     run_time = 1
        # )

        # anim1 = text.animate.rotate(90 * DEGREES).set_run_time(2)
        # anim2 = text.animate.scale(2).set_run_time(1)
        # self.play(
        #     AnimationGroup([anim1, anim2]),
        #     run_time = 2
        # )

        # # TODO: This plays two animations consecutively
        # # Simple form
        # self.play(text.animate.scale(0.5), run_time = 0.5)
        # self.play(text.animate.scale(2), run_time = 0.5)

        # # TODO: This plays two animations consecutively
        # # Succession form
        # # TODO: Can I make that the succession, that is supposed
        # # to last 4 seconds, to last only 2 by passing a run_time
        # # to the 'self.play()' because the other run_times are
        # # updated according to the general one (?) That would be
        # # awesome
        # self.play(
        #     Succession(
        #         ApplyMethod(text.scale, 0.5, run_time = 15 / 60),
        #         ApplyMethod(text.scale, 2, run_time = 45 / 60),
        #     ),
        #     run_time = 4
        # )

        # # TODO: This plays two animations at the same one, and
        # # the first one is two animations consecutively
        # self.play(
        #     AnimationGroup(
        #         Succession(
        #             ApplyMethod(text.scale, 0.5, run_time = 2),
        #             ApplyMethod(text.scale, 2, run_time = 2),
        #         ),
        #         ApplyMethod(text.set_color, RED, run_time = 1)
        #         # Succession(
        #         #     #ApplyMethod(text.set_color(BLUE), run_time = 1 / 60),
        #         #     # TODO: How can I make this wait between animations (?)
        #         #     #ApplyMethod(self.wait, 1),
        #         #     #Wait(1), # This blocks the other ones
        #         #     ApplyMethod(text.set_color, RED, run_time = 2) # This blocks the previous succession
        #         # )
        #     )
        # )

        # TODO: I want to be able to do this below
        # What if I want to use animations like:
        # mobj1.animate.X  -> from 0 to 1 do this, then from 1 to 2 wait
        # mobj2.animate.Y  -> from 0 to 1 wait, from 1 to 2 do this

        """self.play(
            AnimationGroup([
                Succession(
                    ApplyMethod(text.scale, 2),
                    ApplyMethod(text.scale, 0.5),
                ),
                text.animate.rotate(90 * DEGREES)
            ],
            run_time = 2),
            rate_func = linear
        )"""

        # die1 = RoundedRectangle(height=1, width=1, corner_radius=0.2).move_to([0, 0, 0])
        # die1.save_state()
        # def update_die1(mob, alpha):
        #     mob.restore()
        #     mob.become(
        #         RoundedRectangle(
        #             height=1,
        #             width=1,
        #             corner_radius=0.2,
        #         ).move_to([interpolate(0, 1.9, alpha), 0, 0]).rotate(interpolate(0, 90*DEGREES, alpha))
        #     )
        # self.play(
        #     UpdateFromAlphaFunc(die1, update_die1)
        # )
        # self.wait()

        # self.play(
        #     AnimationGroup(
        #         [
        #             text.animate.shift(UP),
        #             text.animate.rotate(PI / 8),
        #             text.animate.scale(2)
        #         ]
        #     ), 
        #     run_time = 2
        # )

        # self.wait(1)


"""
# Very interesting (https://github.com/m-bain/whisperX) to get timestamps of each word
# But the one I use is this one (https://github.com/linto-ai/whisper-timestamped)
"""
from yta_general_utils.file.checker import FileValidator
from yta_general_utils.programming.enum import YTAEnum as Enum
from faster_whisper import WhisperModel
from typing import Union

import whisper_timestamped


class WhisperModel(Enum):
    """
    The model of Whisper you want to use to detect
    the audio.

    See more:
    https://github.com/openai/whisper?tab=readme-ov-file#available-models-and-languages
    """
    TINY = 'tiny'
    """
    Trained with 39M parameters.
    Required VRAM: ~1GB.
    Relative speed: ~10x.
    """
    BASE = 'base'
    """
    Trained with 74M parameters.
    Required VRAM: ~1GB.
    Relative speed: ~7x.
    """
    SMALL = 'small'
    """
    Trained with 244M parameters.
    Required VRAM: ~2GB.
    Relative speed: ~4x.
    """
    MEDIUM = 'medium'
    """
    Trained with 769M parameters.
    Required VRAM: ~5GB.
    Relative speed: ~2x.
    """
    LARGE = 'large'
    """
    Trained with 1550M parameters.
    Required VRAM: ~10GB.
    Relative speed: ~1x.
    """
    TURBO = 'turbo'
    """
    Trained with 809M parameters.
    Required VRAM: ~6GB.
    Relative speed: ~8x.
    """
class WhisperTranscriptor:
    """
    An audio transcriptor based on the Whisper
    model that is capable of transcribing audios
    giving the start and end time of each word
    said in the audio.
    """
    @staticmethod
    def transcribe_with_timestamps(audio_filename: str, model: WhisperModel = WhisperModel.SMALL, initial_prompt: Union[str, None] = None):
        """
        Transcribe the provided 'audio_filename' using the
        specified 'model' and obtains a list of 'words' (with
        the 'start' and 'end' times) and the whole 'text' in
        the audio.

        This method returns the tuple (words, text).
        """
        if not FileValidator.file_is_audio_file(audio_filename):
            raise Exception('The provided "audio_filename" is not a valid audio file.')
        
        model = WhisperModel.to_enum(model)

        audio = whisper_timestamped.load_audio(audio_filename)
        model = whisper_timestamped.load_model(model.value, device = 'cpu')

        # See this: https://github.com/openai/openai-cookbook/blob/main/examples/Whisper_prompting_guide.ipynb
        # I do this 'initial_prompt' formatting due to some issues when using it 
        # as it was. I read in Cookguide to pass natural sentences like this below
        # and it seems to be working well, so here it is:
        if initial_prompt is not None:
            #initial_prompt = '""""' + initial_prompt + '""""'
            #initial_prompt = 'I know exactly what is said in this audio and I will give it to you (between double quotes) to give me the exact transcription. The audio says, exactly """"' + initial_prompt + '""""'
            initial_prompt = 'I will give you exactly what the audio says (the output), so please ensure it fits. The output must be """"' + initial_prompt + '""""'

        # 'vad' = True would remove silent parts to decrease hallucinations
        # 'detect_disfluences' detects corrections, repetitions, etc. so the
        # word prediction should be more accurate. Useful for natural narrations
        transcription = whisper_timestamped.transcribe(model, audio, language = "es", initial_prompt = initial_prompt)
        """
        'text', which includes the whole text
        'segments', which has the different segments
            'words', inside 'segments', which contains each 
            word and its 'text', 'start' and 'end'
        """
        words = [word for segment in transcription['segments'] for word in segment['words']]
        text = ' '.join([word['text'] for word in words])

        return words, text
    
    @staticmethod
    def transcribe(audio_filename: str, initial_prompt: Union[str, None] = None):
        """
        Obtain the transcription with no timestamps, fast and
        ideal to get ideas from the text, summarize it, etc.

        I recommend you using the 'transcribe_with_timestamps'
        because the result is more complete.

        TODO: I don't know if this method is useful, but I 
        keep it here by now until I decide to remove it or to
        actually keep it definitely
        """
        # TODO: What if 'whisper_timestamped' is better? We should use it
        # TODO: Create FasterWhisperModel Enum and add to params
        model = WhisperModel('base')

        if initial_prompt is not None:
            #initial_prompt = '""""' + initial_prompt + '""""'
            #initial_prompt = 'I know exactly what is said in this audio and I will give it to you (between double quotes) to give me the exact transcription. The audio says, exactly """"' + initial_prompt + '""""'
            initial_prompt = 'I will give you exactly what the audio says (the output), so please ensure it fits. The output must be """"' + initial_prompt + '""""'

        segments, _ = model.transcribe(audio_filename, language = 'es', beam_size = 5, initial_prompt = initial_prompt)

        text = ' '.join([segment.text for segment in segments]).strip()

        return text


from yta_multimedia.video.generation.manim.classes.base_manim_animation import BaseManimAnimation
from yta_multimedia.video.generation.manim.enums import MANIM_RENDERER
from yta_multimedia.video.generation.manim.classes.base_manim_animation_wrapper import BaseManimAnimationWrapper, ManimAnimationType
from yta_multimedia.video.generation.manim.utils.dimensions import fitting_text, ManimDimensions
from yta_multimedia.video.consts import MANIM_SCENE_DEFAULT_SIZE
from yta_general_utils.programming.parameter_validator import NumberValidator
from manim import *
from typing import Union


class TextWordByWordManimAnimationWrapper(BaseManimAnimationWrapper):
    """
    The provided 'text' is shown word by word in the center of the scene
    with a fixed width.
    """
    text: str = None
    duration: float = None

    def __init__(self, text: str, duration: float):
        # TODO: This parameter validation could be done
        # by our specific parameter validator (under
        # construction yet)
        exception_messages = []

        if not text:
            exception_messages.append('No "text" parameter provided.')
        
        if not NumberValidator.is_positive_number(duration) or not NumberValidator.is_number_between(duration, 0, 100, do_include_lower_limit = False):
            exception_messages.append('The "duration" parameter provided is not a positive number between (0, 100] (zero is not valid).')

        if len(exception_messages) > 0:
            raise Exception('\n'.join([exception_message for exception_message in exception_messages]))
        
        self.text = text
        self.duration = duration
        super().__init__(TextWordByWordManimAnimationGenerator, ManimAnimationType.TEXT_ALPHA)

class TextWordByWordManimAnimationGenerator(BaseManimAnimation):
    """
    The provided 'text' is shown word by word in the center of the scene
    with a fixed width.
    """
    def construct(self):
        """
        This method is called by manim when executed by shell and
        will call the scene animation render method to be processed
        and generated.
        """
        self.animate()

    def generate(self, parameters: dict, output_filename: Union[str, None] = None):
        """
        Build the manim animation video and stores it
        locally as the provided 'output_filename', 
        returning the final output filename.

        This method will call the '.animate()' method
        to build the scene with the instructions that
        are written in that method.
        """
        return super().generate(
            parameters,
            renderer = MANIM_RENDERER.CAIRO,
            output_filename = output_filename
        )
    
    def animate(self):
        words = self.parameters['text'].split(' ')
        word_duration = float(self.parameters['duration']) / len(words)
        for word in words:
            text = fitting_text(word, ManimDimensions.manim_width_to_width(MANIM_SCENE_DEFAULT_SIZE[0] / 6))
            text = Text(word, font_size = text.font_size, stroke_width = 2.0, font = 'Arial').shift(DOWN * 0)
            self.add(text)
            self.wait(word_duration)
            self.remove(text)


from yta_multimedia.video.generation.manim.classes.base_manim_animation import BaseManimAnimation
from yta_multimedia.video.generation.manim.classes.base_three_d_manim_animation import BaseThreeDManimAnimation
from yta_general_utils.programming.parameter_validator import PythonValidator
from yta_general_utils.programming.attribute_obtainer import AttributeObtainer
from yta_general_utils.programming.enum import YTAEnum as Enum
from typing import Union


class ManimAnimationType(Enum):
    GENERAL = 'general'
    """
    Manim animation which doesn't have an specific
    type by now, so it is classified as a general
    animation.
    """
    TEXT_ALPHA = 'text_alpha'
    """
    Manim animation which is text with a transparent 
    background, intented to be used as a main video
    overlay.
    """

class BaseManimAnimationWrapper:
    """
    Base class for all the manim animation generator
    classes that we want to have in our system.

    This wrapper is to define the attributes that 
    are needed and the manim animation generator
    class that will be used to generate it.
    """
    types: list[ManimAnimationType] = None
    animation_generator_instance: Union[BaseManimAnimation, BaseThreeDManimAnimation] = None

    def __init__(self, animation_generator_instance_or_class: Union[BaseManimAnimation, BaseThreeDManimAnimation], types: Union[list[ManimAnimationType], ManimAnimationType] = ManimAnimationType.GENERAL):
        if not PythonValidator.is_subclass(animation_generator_instance_or_class, BaseManimAnimation) and not PythonValidator.is_subclass(animation_generator_instance_or_class, BaseThreeDManimAnimation):
            raise Exception('The provided "animation_generator_instance_or_class" is not a subclass nor an instance of a subclass of BaseManimAnimation or BaseThreeDManimAnimation classes.')
        
        if PythonValidator.is_a_class(animation_generator_instance_or_class):
            animation_generator_instance_or_class = animation_generator_instance_or_class()

        types = [
            ManimAnimationType.to_enum(type)
            for type in (types if PythonValidator.is_list(types) else [types if types is not None else ManimAnimationType.GENERAL])
        ]

        self.animation_generator_instance = animation_generator_instance_or_class
        self.types = types

    @property
    def attributes(self):
        """
        Only the values that are actually set on the
        instance are obtained with 'vars'. If you set
        'var_name = None' but you don't do 
        'self.var_name = 33' in the '__init__' method,
        it won't be returned by the 'vars()' method.
        """
        return AttributeObtainer.get_attributes_from_instance(self, ['animation_generator_instance', 'attributes'])

    def generate(self):
        """
        Generate the manim animation if the parameters are
        valid and returns the filename of the generated
        video to be used in the app (you should handle it
        with a 'VideoFileClip(o, has_mask = True)' to load
        it with mask and to be able to handle it).
        """
        return self.animation_generator_instance.generate(self.attributes, output_filename = 'output.mov')


"""
All this file is currently in 'yta_multimedia' library
because I'm testing it, but the concepts contained in
this file don't belong to this library because they are
related to a VideoEditor concept, not to image or video
simple editing.

We need to handle, when working in building a whole video
project, videos as SubClips so we handle all attributes
and, if we subclip a SubClip instance, we .copy() the
previous attributes to the left, center and right clips
we obtain when subclipping. This would preserve previous
configurations and let us manage all the clips, so we
work on top of moviepy library in any change we process
and use moviepy only for basic and frame transformations.

TODO: These classes below will be moved in a near future
to its own project or to 'youtube_autonomous'.
"""
from yta_multimedia.video.parser import VideoParser
from yta_multimedia.image.edition.image_editor import ImageEditor
from yta_multimedia.video.edition.effect.moviepy.mask import ClipGenerator
from yta_multimedia.video.edition.settings import VOLUME_LIMIT, ZOOM_LIMIT, LAYERS_INDEXES_LIMIT, COLOR_TEMPERATURE_LIMIT, MAX_TIMELINE_LAYER_DURATION, BRIGHTNESS_LIMIT, CONTRAST_LIMIT, WHITE_BALANCE_LIMIT
from yta_multimedia.video.edition.effect.m_effect import MEffect
from yta_general_utils.programming.parameter_validator import NumberValidator, PythonValidator
from yta_general_utils.programming.enum import YTAEnum as Enum
from moviepy.Clip import Clip
from moviepy import CompositeVideoClip, concatenate_videoclips
from typing import Union


END_OF_CLIP = 999999
END_OF_TIMELINE = 120
"""
The limit of the timeline length. It is not possible
to generate a project which length is larger than 
this value.

TODO: This value could be very low in alpha and beta
versions for testing.
"""

# TODO: Move this interval generator to 'yta_general_utils'


# TODO: Move this decorator below to another place (?)
def unset_video_processed(func):
    """
    Decorator function that sets the '_video_processed'
    attribute to None to indicate that it has been
    modified and it must be processed again.
    """
    def wrapper(self, value):
        value = func(self, value)
        self._video_processed = None
        
        return value
    
    return wrapper


from yta_general_utils.math.rate_functions import mRateFunction
from yta_general_utils.math.progression import Progression

class VideoSetting:
    """
    Class to represent a video setting to be able to handle
    dynamic setting values and not only simple values. This
    means we can make a video go from 0 contrast to 10 
    contrast increasing it smoothly (for example: 0, 2, 4,
    6, 8 and 10) and not only abruptly (from 0 in one frame
    to 10 in the next frame).
    """
    key: str = None
    initial_value: float = None
    final_value: float = None
    rate_function: mRateFunction = None
    _values: list[float] = None

    def __init__(self, initial_value: float, final_value: float, rate_function: mRateFunction = mRateFunction.LINEAR):
        # TODO: Maybe validate something? I don't know the
        # limits because each setting is different, but by
        # now I'm verifying the 'initial_value' and the
        # 'final_value' when using them on a SubClip
        self.initial_value = initial_value
        self.final_value = final_value
        self.rate_function = rate_function

    def get_values(self, steps: int):
        """
        Obtain an array with the values between the 'initial_value'
        and the 'final_value' according to the 'rate_function'.

        The 'steps' parameter must be the amount of frames in which
        we are planning to apply this setting so we are able to read
        the value for each frame according to its index.
        """
        if self._values is None:
            self._values = Progression(self.initial_value, self.final_value, steps, self.rate_function).values
            # TODO: If we make any method to update values we need
            # to handle a flag if these '_values' have to be
            # recalculated

        return self._values

class SubClip:
    """
    Class to represent a subclip of a clip in which we
    can apply different modifications such as color
    temperature, zoom, movement, etc.

    This class represent the same as one of the subclips
    in any of your video editor apps.
    """
    # TODO: This, as the original video clip (moviepy),
    # maybe must be private to avoid using it directly
    video: Clip = None
    _video_processed: Clip = None
    """
    The video once it's been processed with all its
    attributes and effects.
    """
    # TODO: Add all needed attributes
    # Volume
    _volume: int = None
    # Color attributes
    _color_temperature: VideoSetting = None
    _brightness: int = None
    _contrast: int = None
    _sharpness: int = None
    _white_balance: int = None
    # Zoom and movement
    _zoom: int = None
    x_movement: int = None
    y_movement: int = None
    rotation: int = None
    # Custom effects
    _effects: list[MEffect] = None

    @staticmethod
    def init(video: Clip, start_time: Union[float, None] = 0, end_time: Union[float, None] = END_OF_CLIP):
        """
        This is the only method we need to use when instantiating
        a SubClip so we can obtain the left and right clip result
        of the subclipping process and also the new SubClip
        instance.

        This method returns a tuple with 3 elements, which are the
        left part of the subclipped video, the center part and the
        right part as SubClip instances. If no left or right part,
        they are return as None. So, the possibilities are (from
        left to right):

        - SubClip, SubClip, SubClip
        - SubClip, SubClip, None
        - None, SubClip, SubClip
        - None, SubClip, SubClip
        """
        video = VideoParser.to_moviepy(video)

        left_clip, center_clip, right_clip = subclip_video(video, start_time, end_time)

        return (
            SubClip(left_clip) if left_clip is not None else None,
            SubClip(center_clip),
            SubClip(right_clip) if right_clip is not None else None
        )

    def __init__(self, video: Clip):
        """
        DO NOT USE THIS METHOD DIRECTLY. Use static '.init()'
        method instead.

        The SubClip instantiating method has to be called only
        in the static 'init' method of this class so we are able
        to handle the rest of the clips (if existing) according
        to the subclipping process we do.
        """
        self.video = VideoParser.to_moviepy(video)
        self.volume = 100

    @property
    def video_processed(self):
        if not self._video_processed:
            self._video_processed = self._process()

        return self._video_processed

    @property
    def duration(self):
        """
        Shortcut to the actual duration of the video once
        it's been processed.
        """
        return self.video_processed.duration
    
    @property
    def size(self):
        """
        Shortcut to the actual size of the video once
        it's been processed.
        """
        return self.video_processed.size
    
    @property
    def zoom(self):
        return self._zoom
    
    @zoom.setter
    @unset_video_processed
    def zoom(self, value):
        _validate_zoom(value)
        
        self._zoom = int(value)

    @property
    def color_temperature(self):
        return self._color_temperature
    
    @color_temperature.setter
    @unset_video_processed
    def color_temperature(self, value: VideoSetting):
        _validate_color_temperature(value)

        self._color_temperature = value

    @property
    def brightness(self):
        return self._brightness

    @brightness.setter
    @unset_video_processed
    def brightness(self, value):
        _validate_brightness(value)

        self._brightness = value

    @property
    def contrast(self):
        return self._contrast
    
    @contrast.setter
    @unset_video_processed
    def contrast(self, value):
        _validate_contrast(value)

        self._contrast = value

    @property
    def sharpness(self):
        return self._sharpness
    
    @sharpness.setter
    @unset_video_processed
    def sharpness(self, value):
        _validate_contrast(value)

        self._contrast = value

    @property
    def white_balance(self):
        return self._white_balance
    
    @white_balance.setter
    @unset_video_processed
    def white_balance(self, value):
        _validate_white_balance(value)

        self._white_balance = value

    @property
    def volume(self):
        return self._volume
    
    @volume.setter
    @unset_video_processed
    def volume(self, value):
        _validate_volume(value)
        
        self._volume = int(value)

    @property
    def effects(self):
        return self._effects
    
    @effects.setter
    @unset_video_processed
    def effects(self, value):
        """
        Add a list of custom effects instances to apply on the clip.

        TODO: Is this interesting, or maybe just a method to add them
        one by one (?)
        """
        # TODO: Do we validate anything else than below (?)
        #_validate_effect(value)

        if PythonValidator.is_list(value) and any(not PythonValidator.is_subclass(value, MEffect)):
            raise Exception('Some of the given effects in "value" parameter are not MEffect subclass instances.')

        self._effects = value

    def add_effect(self, effect: MEffect):
        """
        Add the provided 'effect' instance to be applied on the clip.
        """
        if not PythonValidator.is_an_instance(effect) or not PythonValidator.is_subclass(effect, MEffect):
            raise Exception('The provided "effect" parameter is not an instance of a MEffect subclass.')
        
        # TODO: Check that effect is valid (times are valid, there is
        # not another effect that makes it incompatible, etc.)
        self._effects.append(effect)
    
    def _process(self):
        """
        Process the video clip with the attributes set and 
        obtain a copy of the original video clip with those
        attributes and effects applied on it. This method
        uses a black (but transparent) background with the
        same video size to make sure everything works 
        properly.

        This method doesn't change the original clip, it
        applies the changes on a copy of the original one
        and returns that copy modified.
        """
        video = self.video.copy()
        black_background_video = ClipGenerator.get_default_background_video(duration = video.duration)

        from yta_multimedia.video import MPVideo

        # I use this class to ensure it fits the real number
        # of frames so we can modify it correctly
        video_handler = MPVideo(video)

        # TODO: The 3 customizable elements
        # This can be a factor or an absolute value, but I
        # recommend being a factor because it is easier
        resizes = [1 for _ in video_handler.frames_time_moments]
        rotations = [0 for _ in video_handler.frames_time_moments]
        # Position must be, at first, a center position, and then
        # transformed into upper left corner position to be 
        # positioned
        positions = [('center', 'center') for _ in video_handler.frames_time_moments]

        # Functions that need to be processed frame by frame
        def modify_video_frame_by_frame(get_frame, t):
            """
            Modificate anything related to pixel colors, distortion,
            etc. and not to position, rotation or frame size.
            """
            frame = get_frame(t)
            frame_index = video_handler.frame_time_to_frame_index(t, video.fps)

            frame = ImageEditor.modify_color_temperature(frame, self.color_temperature.values[frame_index]) if self.color_temperature is not None else frame
            frame = ImageEditor.modify_brightness(frame, self.brightness) if self.brightness is not None else frame
            frame = ImageEditor.modify_contrast(frame, self.contrast) if self.contrast is not None else frame
            frame = ImageEditor.modify_sharpness(frame, self.sharpness) if self.sharpness is not None else frame
            frame = ImageEditor.modify_white_balance(frame, self.white_balance) if self.white_balance is not None else frame

            return frame
        
        # Apply frame by frame video modifications
        video = video.transform(lambda get_frame, t: modify_video_frame_by_frame(get_frame, t))

        # Functions that can be processed in the whole clip
        # TODO: Careful. Some of these modification can be
        # done before the video positioning or resizing, so
        # the behaviour could be unexpected.
        size = video.size
        if self.zoom is not None:
            size = (self.zoom / 100 * size[0], self.zoom / 100 * size[1])
        
            video = video.resized(size)
            # This is to force the size to be 1920x1080 instead
            # of other bigger values, but thiss does not make 
            # the clip render faster
            # from moviepy.video.fx.Crop import Crop
            # video = Crop(x_center = video.size[0] / 2, y_center = video.size[1] / 2, width = 1920, height = 1080).apply(video)

        # Apply position, rotation and resize frame by frame
        video = video.resized(lambda t: resizes[video_handler.frame_time_to_frame_index(t, video_handler.fps)])
        video = video.with_position(lambda t: positions[video_handler.frame_time_to_frame_index(t, video_handler.fps)])
        # TODO: There is a Rotate Effect for rotating... so we could
        # use it instead of this (?)
        video = video.rotated(lambda t: rotations[video_handler.frame_time_to_frame_index(t, video_handler.fps)])

        # Functions that changes the audio
        if self.volume != 100:
            video = video.with_volume_scaled(self.volume / 100)

        # TODO: This below is repeated in VideoEditor class as
        # '._overlay_video()'
        return CompositeVideoClip([
            black_background_video,
            video
        ])#.with_audio(VideoAudioCombinator(audio_mode).process_audio(background_video, video))

class SubClipOnTimelineLayer:
    """
    Class to represent one of our SubClips but in
    the general project timeline and in a specific
    layer of it, with the start and end moment in
    that timeline, and also the layer in which it
    is placed.

    TODO: This is a concept in test phase. SubClip
    is a more definitive concept.
    """
    subclip: SubClip = None
    start_time: float = None
    """
    The start time on the general project timeline. Do 
    not confuse this term with the start time of a
    moviepy clip.
    """

    @property
    def video_processed(self):
        return self.subclip.video_processed

    @property
    def duration(self):
        """
        Shortcut to the actual duration of the video once
        it's been processed.
        """
        return self.subclip.duration
    
    @property
    def size(self):
        """
        Shortcut to the actual size of the video once it's
        been processed.
        """
        return self.subclip.size
    
    @property
    def end_time(self):
        """
        The end moment on the timeline, based on this
        instance 'start_time' and the real video 'duration'.
        """
        return self.start_time + self.duration

    def __init__(self, subclip: SubClip, start_time: float):
        if not PythonValidator.is_instance(subclip, SubClip):
            raise Exception('The provided "subclip" parameter is not a valid SubClip instance.')
        
        if not NumberValidator.is_number_between(start_time, 0, END_OF_TIMELINE):
            raise Exception(f'The provided "start_time" parameter is not a valid number between in the range (0, {END_OF_TIMELINE})')
        
        self.subclip = subclip
        self.start_time = start_time

    # def _process(self):
    #     return self.subclip._process()

class TimelineLayerType(Enum):
    """
    The type of a timeline layer, which will determine
    which kind of SubClips are accepted by the layer
    with this type.
    """
    VIDEO = 'video'
    """
    The type of layer that only accept video SubClips.
    """
    AUDIO = 'audio'
    """
    The type of layer that only accept audio SubClips.
    """
    # TODO: Probably add 'GreenscreenLayer',
    # 'AlphascreenLayer', 'TextLayer', 'SubtitleLayer',
    # and all needed in a future when this concept 
    # evolves properly

class TimelineLayer:
    index: int = None
    type: TimelineLayerType = None
    _subclips: list[SubClipOnTimelineLayer] = None

    @property
    def subclips(self):
        """
        The list of subclips in this timeline layer, ordered
        according to its 'start_time' from first to be displayed
        to last ones.
        """
        return sorted(self._subclips, key = lambda subclip: subclip.start_time)
    
    @property
    def duration(self):
        """
        The duration of this timeline layer, which is the
        'end_time' of the last displayed subclip, or 0 if no
        clips.
        """
        return self.subclips[-1].end_time if self.subclips else 0
    
    @property
    def all_clips(self):
        """
        The list of all clips needed to fulfill the timeline
        layer completely. This involves the actual subclips
        but also the needed black background clips to put in
        the gaps between the subclips.
        """
        all_subclips = []
        current_time_moment = 0
        # TODO: This 'timeline_duration' must be calculated
        # according to all timeline layers, so the longest
        # one is the one we need to assign here
        timeline_duration = self.subclips[-1].end_time

        for subclip in self.subclips:
            # We fulfill the gap if existing
            if subclip.start_time > current_time_moment:
                all_subclips.append(ClipGenerator.get_default_background_video(subclip.size, subclip.start_time - current_time_moment))
            
            # We add the existing subclip
            all_subclips.append(subclip.video_processed)
            
            current_time_moment = subclip.end_time
        
        # Check if gap at the end due to other longer layers
        if current_time_moment < timeline_duration:
            all_subclips.append(ClipGenerator.get_default_background_video(subclip.size, timeline_duration - current_time_moment))
        
        return all_subclips

    def __init__(self, index: int = 0, type: TimelineLayerType = TimelineLayerType.VIDEO):
        _validate_layer_index(index)
        type = TimelineLayerType.to_enum(type) if type is not None else TimelineLayerType.VIDEO

        self.index = index
        self.type = type
        self._subclips = []

    def add_subclip(self, subclip: SubClip, start_time: float):
        """
        Append the provided 'subclip' at the end of the list.

        TODO: Being in the end of the list doesn't mean being
        the last one displayed. By now I'm storing them just
        one after another and ordering them when trying to get
        them as a property. This will change in a near future
        to be more eficient.
        """
        if not PythonValidator.is_instance(subclip, SubClip):
            raise Exception('The provided "subclip" parameter is not a valid instance of the SubClip class.')
        
        if not NumberValidator.is_number_between(start_time, 0, MAX_TIMELINE_LAYER_DURATION):
            raise Exception('The provided "start_time" is not a valid value')

        # 
        # TODO: Check that the 'start_time' or the 'duration'
        # doesn't collide with another existing subclip. If
        # yes, choose what strategy to follow
        if any(subclip.start_time <= start_time <= subclip.end_time for subclip in self.subclips):
            raise Exception(f'There is one existing subclip at the {str(start_time)} time position.')

        self._subclips.append(SubClipOnTimelineLayer(subclip, start_time))

    def remove_subclip(self, index: int):
        """
        Delete the subclip in the provided 'index' position of
        the list (if existing), or raises an Exception if it
        doesn't exist or the list is empty.
        """
        # TODO: Maybe remove by passing the instance (?)
        if PythonValidator.is_empty_list(self._subclips):
            # TODO: Maybe I should not raise an Exception here...
            raise Exception('No subclips to remove.')
        
        if not NumberValidator.is_number_between(0, len(self._subclips)):
            raise Exception(f'The provided "index" is not a valid index (must be between [0, {str(len(self._subclips))}).')

        # TODO: Be very careful, I have a 'subclips' property which
        # returns the subclips ordered, but the raw '_subclips'
        # property is not ordered, so here the 'index' is potentially
        # wrong. Think how to handle this in a near future, please.
        del self._subclips[index]

    def build(self):
        """
        Concatenate all the timeline layer subclips (fulfilling the
        gaps with black transparent background clips) and return the
        concatenated clip.
        """
        # TODO: What if I have one non-transparent and transparent
        # clips in this timeline layer? They will be treated in a
        # similar way so it is not the expected behaviour...
        return concatenate_videoclips(self.all_clips)
    
class VideoProject:
    """
    Class representing a whole but single video project
    in which we have a timeline with different layers
    and clips on them.
    """
    timeline_layers: list[TimelineLayer] = None
    screen_size: tuple[int, int] = None
    """
    Dimensions of the final video that has to be
    exported.
    """

    def __init__(self, screen_size: tuple[int, int] = (1920, 1080)):
        self.timeline_layers = [
            TimelineLayer(0, TimelineLayerType.VIDEO),
            # TODO: I Simplify everything by now and I only
            # handle one video layer
            #TimelineLayer(0, TimelineLayerType.AUDIO)
        ]
        # TODO: Validate 'screen_size'
        self.screen_size = screen_size

    def get_layers(self, type: TimelineLayerType) -> list[TimelineLayer]:
        """
        Get the timeline layers of the provided 'type' sorted by
        index in ascending order.
        """
        type = TimelineLayerType.to_enum(type)

        return sorted(
            [layer for layer in self.timeline_layers if layer.type == type], 
            key = lambda layer: layer.index
        )
    
    def get_last_layer_index(self, type: TimelineLayerType) -> Union[int, None]:
        """
        Get the last index used for the layers of the given
        'type', that will be None if no layers of that type.
        """
        type = TimelineLayerType.to_enum(type)

        layers = self.get_layers(type)

        return layers[-1:].index if not PythonValidator.is_empty_list(layers) else None

    def add_layer(self, type: TimelineLayerType = TimelineLayerType.VIDEO) -> int:
        """
        Add a new layer of the provided 'type' and returns
        the index in which it has been placed.
        """
        type = TimelineLayerType.to_enum(type) if type is not None else TimelineLayerType.VIDEO

        layers = self.get_layers(type)
        index = len(layers) + 1 if layers else 0

        self.timeline_layers.append(TimelineLayer(index, type))

        return index
    
    def remove_layer(self, layer: TimelineLayer):
        if not PythonValidator.is_instance(layer, TimelineLayer):
            raise Exception('The provided "layer" is not an instance of the TimelineLayer class.')
        
        if layer not in self.timeline_layers:
            raise Exception('The provided "layer" does not exist in this project.')

        self.timeline_layers.remove(layer)

    def build(self):
        from yta_multimedia.video.edition.duration import set_video_duration, ExtendVideoMode
        # TODO: Remove this code below when this is done 
        # individually by each layer. By now I'm forcing
        # all layers clips to have the same duration as
        # the longest layer clip has, but must be done
        # using a general and common 'max_duration'
        # property that is not calculated here.
        layers_clips = [
            layer.build()
            for layer in self.timeline_layers
        ]
        # TODO: Omg, cyclic import issue here again...
        max_duration = max(layer_clip.duration for layer_clip in layers_clips)
        layers_clips = [
            set_video_duration(layer_clip, max_duration, extend_mode = ExtendVideoMode.BLACK_TRANSPARENT_BACKGROUND)
            for layer_clip in layers_clips
        ]

        return CompositeVideoClip(layers_clips)





def subclip_video(video: Clip, start_time: float, end_time: float) -> tuple[Union[Clip, None], Clip, Union[Clip, None]]:
    """
    Subclip the provided 'video' into 3 different subclips,
    according to the provided 'start_time' and 'end_time',
    and return them as a tuple of those 3 clips. First and
    third clip could be None.

    The first clip will be None when 'start_time' is 0, and 
    the third one when the 'end_time' is equal to the given
    'video' duration.
    """
    video = VideoParser.to_moviepy(video)

    left = None if (start_time == 0 or start_time == None) else video.with_subclip(0, start_time)
    center = video.with_subclip(start_time, end_time)
    right = None if (end_time is None or end_time >= video.duration) else video.with_subclip(start_time = end_time)

    return left, center, right



# TODO: This must be maybe moved to another file
# because it is mixed with the 'subclip_video'
# method...
def _validate_layer_index(layer_index: int):
    if not NumberValidator.is_number_between(layer_index, LAYERS_INDEXES_LIMIT[0], LAYERS_INDEXES_LIMIT[1]):
        raise Exception(f'The provided "layer_index" is not a valid layer, it must be an int value between [{LAYERS_INDEXES_LIMIT[0]}, {LAYERS_INDEXES_LIMIT[1]}].')
    
def _validate_zoom(zoom: int):
    if not NumberValidator.is_number_between(zoom, ZOOM_LIMIT[0], ZOOM_LIMIT[1]):
        raise Exception(f'The "zoom" parameter provided is not a number between [{ZOOM_LIMIT[0]}, {ZOOM_LIMIT[1]}].')
    
def _validate_color_temperature(color_temperature: VideoSetting):
    if not PythonValidator.is_instance(color_temperature, VideoSetting):
        raise Exception('The provided "color_temperature" is not a VideoSetting instance.')
    
    if not NumberValidator.is_number_between(color_temperature.initial_value, COLOR_TEMPERATURE_LIMIT[0], COLOR_TEMPERATURE_LIMIT[1]):
        raise Exception(f'The "color_temperature" parameter provided "initial_value" is not a number between [{COLOR_TEMPERATURE_LIMIT[0]}, {COLOR_TEMPERATURE_LIMIT[1]}].')

    if not NumberValidator.is_number_between(color_temperature.final_value, COLOR_TEMPERATURE_LIMIT[0], COLOR_TEMPERATURE_LIMIT[1]):
        raise Exception(f'The "color_temperature" parameter provided "final_value" is not a number between [{COLOR_TEMPERATURE_LIMIT[0]}, {COLOR_TEMPERATURE_LIMIT[1]}].')
    
def _validate_brightness(brightness: int):
    if not NumberValidator.is_number_between(brightness, BRIGHTNESS_LIMIT[0], BRIGHTNESS_LIMIT[1]):
        raise Exception(f'The "brightness" parameter provided is not a number between [{BRIGHTNESS_LIMIT[0]}, {BRIGHTNESS_LIMIT[1]}].')

def _validate_contrast(contrast: int):
    if not NumberValidator.is_number_between(contrast, CONTRAST_LIMIT[0], CONTRAST_LIMIT[1]):
        raise Exception(f'The "contrast" parameter provided is not a number between [{CONTRAST_LIMIT[0]}, {CONTRAST_LIMIT[1]}].')
    
def _validate_white_balance(white_balance: int):
    if not NumberValidator.is_number_between(white_balance, WHITE_BALANCE_LIMIT[0], WHITE_BALANCE_LIMIT[1]):
        raise Exception(f'The "white_balance" parameter provided is not a number between [{WHITE_BALANCE_LIMIT[0]}, {WHITE_BALANCE_LIMIT[1]}].')

def _validate_volume(volume: int):
    if not NumberValidator.is_number_between(volume, VOLUME_LIMIT[0], VOLUME_LIMIT[1]):
        raise Exception(f'The "volume" parameter provided is not a number between [{VOLUME_LIMIT[0]}, {VOLUME_LIMIT[1]}].')


"""
    Thanks to https://omaraflak.medium.com/b%C3%A9zier-interpolation-8033e9a262c2
    for interpolation explanation and code.

    Thanks to https://github.com/vmichals/python-algos/blob/master/catmull_rom_spline.py
    for the catmull-rom-spline interpolation code.
"""
from yta_general_utils.programming.parameter_validator import NumberValidator, PythonValidator
from yta_multimedia.video.edition.effect.moviepy.position.objects.coordinate import Coordinate
from yta_multimedia.video.position import Position
from yta_general_utils.programming.enum import YTAEnum as Enum
from yta_general_utils.math import Math
from scipy import interpolate
from typing import Callable, Union

import numpy as np
import matplotlib.pyplot as plt


# Here below are some functions to obtain the graphic
# between points
class InterpolationFormula:
    """
    Class to encapsulate interpolation formulas.
    """
    @staticmethod
    def catmull_rom_spline(spline_x: float, p0, p1, p2, p3):
        """
        Function to evaluate one spline of the Catmull-Rom
        between two points by using 4 control points.

        It returns the x,y point in the spline.

        Computes interpolated y-coord for given x-coord using
        Catmull-Rom. The 'spline_x' is the current position in
        the graphic interpolation between the points p0 and p1
        represented by a value between 0 and 1.

        Computes an interpolated y-coordinate for the given
        x-coordinate between the support points v1 and v2. The
        neighboring support points v0 and v3 are used by
        Catmull-Rom to ensure a smooth transition between the
        spline segments.
        """
        c1 = 1. * p1
        c2 = -.5 * p0 + .5 * p2
        c3 = 1. * p0 + -2.5 * p1 + 2. * p2 -.5 * p3
        c4 = -.5 * p0 + 1.5 * p1 + -1.5 * p2 + .5 * p3

        # TODO: We can apply a RateFunction here
        # in 'spline_x' to set the speed as
        # non-linear
        return (((c4 * spline_x + c3) * spline_x + c2) * spline_x + c1)
    
    @staticmethod
    def linear(spline_x: float, p0, p1):
        """
        Function to evaluate one spline of a linear function
        between two points.

        It returns the x,y point in the spline.

        Computes interpolated y-coord for given x-coord using
        linear function.
        """
        # TODO: We can apply a RateFunction here
        # in 'spline_x' to set the speed as
        # non-linear
        x = p0[0] + (p1[0] - p0[0]) * spline_x
        y = p0[1] + (p1[1] - p0[1]) * spline_x

        return (x, y)

class GraphicRateFunction:
    """
    Class to simplify the way we work with rate functions
    being able to personalize a graphic that will allow us
    using it as a rate function to create our own effects.

    We receive a list of points that are coordinates of the
    graph we want to create as a rate function, and the 
    code will analyze it and obtain the function that allows
    us to determine the value of any 't' value between 0.0
    and 1.0.
    
    The 't' value provided is representing the distance in the
    graphic curve that has been generated with the points
    provided and the interpolation method applied. This curve
    is representing some effect, animation or transition
    duration, from 0 to 1. So, if the provided 't' is 0.20,
    that is representing the moment in which the 20% of the
    animation has been played, and will return the 'y' value
    that must be used in the animation for that specific 
    moment.

    We have a graphic representation of a function in which we
    use the 'x' and 'y' axis to represent it. There is another
    value that we call 't' and it is one point within the 
    graphic that has been generated with the provided points
    and the 'interpolation_formula'. This 't' is representing
    an instant during the execution of an animation or
    transition.
    """
    coordinates: list[Coordinate]
    """
    The list of coordinates to represent the graphic we
    want to represent.
    """
    interpolation_formula: Callable
    """
    The method to calculate the interpolation between each
    pair of points.

    TODO: By now it is the same for the whole graph but it
    could change in a near future and be specific for each
    pair of points. We should create an InterpolationStep
    or similar to include 2 coordinates and 1 interpolation
    formula, and receive a list of instances of that class
    as a parameter to be able to customize each type of
    interpolation.
    """

    def __init__(self, points: list[Coordinate, Position, tuple], interpolation_formula: Callable):
        if not PythonValidator.is_list(points) or len(points) < 4 or any(not PythonValidator.is_instance(point, [Coordinate, Position, tuple]) for point in points):
            raise Exception(f'The "points" parameter must be a list of Coordinate, Position with a minimum of 4 elements.')
    
        if not PythonValidator.is_class_staticmethod(InterpolationFormula, interpolation_formula):
            raise Exception('The provided "interpolation_formula" is not a valid InterpolationFormula class method.')

        # TODO: Maybe if less than 4 elements: add 2 in between those 2
        # TODO: Should I allow this below (?) If I'm trying to obtain a
        # path I think yes... I will be using a 't' that is actually the
        # distance in the graph once it's been built, so I don't care if
        # I am going back or not, because it is like if just have two 
        # points in that moment and we want to know the value in one 
        # specific point of the graph between those 2 points, so I think
        # I should accept going towards or backwards
        # if not all(points[i].x < points[i + 1].x for i in range(len(points) - 1)):
        #     raise Exception('Points must be consecutive, so the "x" value of each point must be greater than the previous one. Please, check the provided "points" parameter.')

        # Turn all points to coordinates, normalize them and
        # obtain the normalized x,y in the format we need
        coords = []
        for point in points:
            coord = point
            if PythonValidator.is_instance(point, Position):
                coord = point.as_video_position().coordinate
            elif PythonValidator.is_tuple(point):
                if len(point) != 2:
                    # TODO: Maybe apply the normalization limits as limits
                    # here for each position tuple element
                    raise Exception('Provided "position" is a tuple but does not have 2 values.')
                point = Coordinate(point[0], point[1])

            coord = coord.normalize()
            coords.append(coord)

        self.coordinates = coords
        self.interpolation_formula = interpolation_formula

    def get_coord(self, t: float):
        """
        Evaluate the function for the provided 'x' value and
        return the graphic coordinate corresponding to that
        't' value, that must be a normalized value (between
        0 and 1, both inclusive) that represents the distance
        in the generated graphic.

        A graphic with 5 coordinates will generate 4 splines
        of the same length. The start of each of those splines
        will be [t = 0, t = 0.25, t = 0.5, t = 0.75]. So, if
        you provide t = 0.20, the returned coordinate will be
        contained in the first spline because it is below 0.25.
        """
        if not NumberValidator.is_number_between(t, 0.0, 1.0):
            raise Exception('The provided "t" parameter "{str(t)}" must be  between 0 and 1, both inclusive.')

        num_of_splines = len(self.coordinates) - 1
        spline_size = 1
        if num_of_splines > 1:
            spline_size = 1 / num_of_splines

        if t == 1.0:
            spline_index = num_of_splines - 1
        else:
            # Check which spline we need to use
            spline_index = int(t // spline_size)

        # If we receive 0.28 => 0.28 - 0.25 = 0.03 / spline_size
        # A distance of 0.03 in the global graph is lower than the
        # real distance in any splice
        # 0.28 = 0.28 % (1 * 0.25) / 0.25 = 0.03 / 0.25 = 0.12
        previous_spline_size = spline_index * spline_size

        if previous_spline_size:
            t = t % (spline_index * spline_size) / spline_size
        else:
            t = t / spline_size

        if PythonValidator.is_class_staticmethod(InterpolationFormula, self.interpolation_formula, InterpolationFormula.catmull_rom_spline.__name__):
            numpy_coords = Coordinate.to_numpy(self.coordinates)

            if spline_index == 0: # First spline
                # We need to estimate the first point
                p0 = numpy_coords[0] - (numpy_coords[1] - numpy_coords[0])
                p1 = numpy_coords[0]
                p2 = numpy_coords[1]
                p3 = numpy_coords[2]
            elif spline_index == (num_of_splines - 1): # Last spline
                p0 = numpy_coords[spline_index - 1]
                p1 = numpy_coords[spline_index]
                p2 = numpy_coords[spline_index + 1]
                # We need to estimate the last point
                p3 = numpy_coords[spline_index + 1] + (numpy_coords[spline_index + 1] - numpy_coords[spline_index])
            else:
                p0 = numpy_coords[spline_index - 1]
                p1 = numpy_coords[spline_index]
                p2 = numpy_coords[spline_index + 1]
                p3 = numpy_coords[spline_index + 2]

            coord = self.interpolation_formula(t, p0, p1, p2, p3)
        elif PythonValidator.is_class_staticmethod(InterpolationFormula, self.interpolation_formula, InterpolationFormula.linear.__name__):
            p0 = numpy_coords[spline_index]
            p1 = numpy_coords[spline_index + 1]

            coord = self.interpolation_formula(t, p0, p1)
        
        # We maybe want the 'y' value as we are representing
        # something in a graphic, but we get the whole coord
        # at this point
        coord = Coordinate(coord[0], coord[1], self.coordinates[0].is_normalized)

        return coord
    
    def get_y(self, t: float):
        """
        Evaluate the function for the provided 't' value and
        return the graphic coordinate 'y' value corresponding
        to that 't' value that represents the distance in the
        generated graphic.

        A graphic with 5 coordinates will generate 4 splines
        of the same length. The start of each of those splines
        will be [t = 0, t = 0.25, t = 0.5, t = 0.75]. So, if
        you provide t = 0.20, the returned coordinate will be
        contained in the first spline because it is below 0.25.
        """
        return self.get_coord(t)[1]

    def plot(self, do_denormalize_values: bool = False):
        """
        This method is just for testing and shows the graphic
        using the matplotlib library.

        If you are plotting a path movement along the screen,
        set the 'do_denormalize_values' as True to make the
        plot show the real movement the video will do through
        the scene. If you are plotting a graph used for any
        other type of effect, keep the value as False.
        """
        if do_denormalize_values:
            # I denormalize values to be able to watch the path
            # movement properly
            self.coordinates = [coordinate.denormalize() for coordinate in self.coordinates]

        graph = np.array([self.get_coord(t).as_tuple() for t in np.linspace(0, 1, len(self.coordinates) * 100)])
        points = np.array([coordinate.as_tuple() for coordinate in self.coordinates])

        plt.plot(graph[:, 0], graph[:, 1], label = 'Graphic representation', color = 'blue')
        plt.scatter(points[:, 0], points[:, 1], color = 'red', label = 'Control points')
        plt.legend()
        plt.title('Graphic representation')
        plt.xlabel('x')
        plt.ylabel('y')

        if do_denormalize_values:
            # I invert the axis to see the graph as the video
            # movement because denormalization if because of
            # this
            plt.gca().invert_yaxis()
            
        plt.grid(True)
        plt.show()

    # TODO: If we want to apply one GraphicRateFunction to some
    # of our moviepy animations, we will be working with a 't'
    # value representing the time of the video, that can easily
    # be greater than 1, so we need to normalize it with the
    # video duration and use that normalized value to provide it
    # to this 'get_y()' or 'get_coord()' method.






class InterpolationMethod(Enum):
    """
    Enum class to include the available interpolation
    methods and to predict an 'y' value according to
    each specific method.

    This Enum class is pretended to be used with the
    InterpolatedPair class.
    """
    LINEAR = 'linear'
    QUADRATIC = 'quadratic'
    """
    This interpolation method needs 3 coordinates and
    x cannot be repeated.
    """
    CUBIC = 'cubic'
    """
    This interpolation method needs 4 coordinates and
    x cannot be repeated.
    """
    # TODO: Explain this above better, please

    def get_y_from_d(self, coordinate_a, coordinate_b, d: float):
        """
        Obtain the 'y' value by providing a 'd' distance that
        is the distance from one coordinate to another measured
        in normalized value between 0.0 and 1.0, representing
        the amount of distance traveled in the interpolation.
        """
        return self.get_coord_from_d(coordinate_a, coordinate_b, d)[1]

    def get_y_from_x(self, coordinate_a, coordinate_b, x: float):
        """
        Obtain the interpolated 'y' value for the given 'x'
        value, based on this interpolation method and the
        also given 'coordinate_a' and 'coordinate_b' coords.

        The 'x' parameter must be a normalized value (between
        0.0 and 1.0), representing an 'x' value in between the
        two provided coordinates.
        """
        return self.get_coord_from_x(coordinate_a, coordinate_b, x)[1]
    
    def get_coord_from_d(self, coordinate_a, coordinate_b, d: float):
        """
        Obtain the coordinate value by providing a 'd' distance
        that is the distance from one coordinate to another
        measured in normalized value between 0.0 and 1.0,
        representing the amount of distance traveled in the
        interpolation.
        """
        # TODO: Turn 'd' into x
        if not NumberValidator.is_number_between(d, 0.0, 1.0):
            raise Exception(f'The provided "d" parameter "{str(d)}" must be a value between 0.0 and 1.0.')
        
        # Calculate the 'x' according to the provided distance
        # 'd' traveled through the interpolation
        x = coordinate_a[0] + d * (coordinate_b[0] - coordinate_a[0])

        return self.get_coord_from_x(coordinate_a, coordinate_b, x)

    def get_coord_from_x(self, coordinate_a, coordinate_b, x: float):
        """
        Obtain the interpolated 'y' value for the given 'x'
        value, based on this interpolation method and the
        also given 'coordinate_a' and 'coordinate_b' coords.

        The 'x' parameter must be a normalized value (between
        0.0 and 1.0), representing an 'x' value in between the
        two provided coordinates.
        """
        # TODO: Move this code to 'get_coord_from_x' and call
        # it from here to obtain the y

        if not PythonValidator.is_tuple(coordinate_a) or len(coordinate_a) != 2 or not PythonValidator.is_tuple(coordinate_b) or len(coordinate_b) != 2:
            raise Exception('The provided "coordinate_a" or "coordinate_b" parameter is not a tuple of (x, y) values.')
        
        # Coordinates must be in order to be able to calculate
        # the interpolation. Don't worry, the interpolation will
        # be the same and the 'y' will be ok
        coordinates_in_order = sorted([coordinate_a, coordinate_b], key = lambda coord: coord[0])
        xs = [coordinates_in_order[0][0], coordinates_in_order[1][0]]
        ys = [coordinates_in_order[0][1], coordinates_in_order[1][1]]

        if not NumberValidator.is_number_between(x, xs[0], xs[1]):
            raise Exception(f'The provided "x" parameter "{str(x)}" is out of bounds (limits are [{str(xs[0])}, {str(xs[1])}]).')

        # TODO: Complete with all the interpolation methods
        if self == InterpolationMethod.LINEAR:
            # This method needs only 2 coordinates
            xs = [coordinate_a[0], coordinate_b[0]]
            ys = [coordinate_a[1], coordinate_b[1]]
            y = interpolate.interp1d(xs, ys, kind = InterpolationMethod.LINEAR.value)(x)
        elif self == InterpolationMethod.QUADRATIC:
            xs = [coordinate_a[0], coordinate_b[0]]
            ys = [coordinate_a[1], coordinate_b[1]]
            # We estimate a point in between the other two points
            xs.append((xs[0] + xs[1]) / 2)
            # We apply some randomness to the y of that point (or
            # it will become linear if we don't)
            # The more the distance between the points it is, the
            # more this value can be.
            randomness_factor = 0.95
            ys.append((ys[0] + ys[1]) / 2 * randomness_factor)
            y = interpolate.interp1d(xs, ys, kind = InterpolationMethod.QUADRATIC.value)(x)
        elif self == InterpolationMethod.CUBIC:
            # This method needs at least 4 coordinates
            #y = interpolate.interp1d(xs, ys, kind = 'linear')(x)
            # There is an 'assume_sorted' param as False, so it is
            # ok like this
            # TODO: Estimate other coords
            x0 = coordinate_a[0] - (coordinate_b[0] - coordinate_a[0])
            y0 = coordinate_a[1] - (coordinate_b[1] - coordinate_a[1])
            x3 = coordinate_b[0] + (coordinate_b[0] - coordinate_a[0])
            y3 = coordinate_b[1] + (coordinate_b[1] - coordinate_a[1])
            xs = [x0, coordinate_a[0], coordinate_b[0], x3]
            # TODO: This value, if random, has to be calculated
            # once for the whole interpolation, not for each y
            # value calculation or it will generate weird graphic
            #randomness_factor = randrangefloat(0.9, 1.1, 0.01)
            d = abs(x - coordinate_a[0]) / abs(coordinate_b[0] - coordinate_a[0])
            randomness_factor = 0.9
            ys = [y0 * randomness_factor, coordinate_a[1], coordinate_b[1], y3 * randomness_factor]
            y = interpolate.interp1d(xs, ys, kind = InterpolationMethod.CUBIC.value)(x)
        else:
            raise Exception(f'The "{self.name}" interpolation method has not been implemented yet.')
        # TODO: Implement more methods
        
        x = float(x)
        y = float(y)

        # TODO: If I invert, do I need to do a rest or something (?)

        return (x, y)
    
class InterpolatedPair:
    """
    A pair of coordinates connected by the provided interpolation
    method.
    """
    coordinate_a: Coordinate
    coordinate_b: Coordinate
    interpolation_method: InterpolationMethod
    _max_x: float = None
    _min_x: float = None

    @property
    def max_x(self):
        if not self._max_x:
            self._max_x = max([self.coordinate_a.x, self.coordinate_b.x], key = lambda coord: coord)

        return self._max_x
    
    @property
    def min_x(self):
        if not self._min_x:
            self._min_x = min([self.coordinate_a.x, self.coordinate_b.x], key = lambda coord: coord)

        return self._min_x

    def __init__(self, coordinate_a: Union[Coordinate, Position], coordinate_b: Union[Coordinate, Position], interpolation_method: InterpolationMethod = InterpolationMethod.LINEAR):
        if not PythonValidator.is_instance(coordinate_a, [Coordinate, Position]) or not PythonValidator.is_instance(coordinate_b, [Coordinate, Position]):
            raise Exception(f'The provided "coordinate_a" or "coordinate_b" is not a valid Coordinate or Position.')
        
        interpolation_method = InterpolationMethod.to_enum(interpolation_method)

        # Turn input parameters to Coordinate instances
        if PythonValidator.is_instance(coordinate_a, Position):
            coordinate_a = coordinate_a.get_moviepy_center_tuple()
            coordinate_a = Coordinate(coordinate_a[0], coordinate_a[1])

        if PythonValidator.is_instance(coordinate_b, Position):
            coordinate_b = coordinate_b.get_moviepy_center_tuple()
            coordinate_b = Coordinate(coordinate_b[0], coordinate_b[1])

        # We store them as normalized coordinates
        self.coordinate_a = coordinate_a.normalize()
        self.coordinate_b = coordinate_b.normalize()
        self.interpolation_method = interpolation_method

    def get_coord_from_x(self, x: float):
        """
        Get the coordinate that is in the provided 'x' 
        position (that must be a value between the two
        coordinate x axis we have).
        """
        # TODO: Maybe handle 'is_normalized' parameter
        # here also
        if not NumberValidator.is_number_between(x, self.min_x, self.max_x):
            raise Exception(f'The provided "d" parameter "{str(x)}" must be a value between 0.0 and 1.0.')
        
        y = self._get_y_from_x(x)

        return (x, y)

    def get_coord_from_distance(self, d: float):
        """
        Get the coordinate from the given distance 'd' that
        is the distance we will use to obtain the point in
        that distance between the two coordinates.
        """
        if not NumberValidator.is_number_between(d, 0.0, 1.0):
            raise Exception(f'The provided "d" parameter "{str(d)}" must be a value between 0.0 and 1.0.')
        
        # Calculate the 'x' according to the provided distance
        # 'd' traveled through the interpolation
        x = self.coordinate_a.x + d * (self.coordinate_b.x - self.coordinate_a.x)
        y = self._get_y_from_x(x)

        return (x, y)

    def _get_y_from_d(self, d: float):
        """
        Obtain the 'y' value by providing a 'd' distance that is
        the distance from one coordinate to another measured in
        normalized value between 0.0 and 1.0, representing the
        amount of distance traveled in the interpolation.
        """
        if not NumberValidator.is_number_between(d, 0.0, 1.0):
            raise Exception(f'The provided "d" parameter "{str(d)}" must be a value between 0.0 and 1.0.')
        
        x = self.coordinate_a.x + d * (self.coordinate_b.x - self.coordinate_a.x)

        return self._get_y_from_x(x)

    def _get_y_from_x(self, x: float):
        """
        Obtain the 'y' value by providing a 'x' value that is the
        x in value and must be in between the two coordinates.
        """
        if not NumberValidator.is_number_between(x, self.min_x, self.max_x):
            raise Exception(f'The provided "x" parameter "{str(x)}" must be a value between 0.0 and 1.0.')
        
        # TODO: What about non-consecutive (?)
        y = self.interpolation_method.get_y_from_x(self.coordinate_a.as_tuple(), self.coordinate_b.as_tuple(), x)

        return y

class GraphicInterpolation:
    """
    A graphic to represent points interpolated and to
    be able to calculate 'y' positions in those
    interpolations.
    """
    pairs_of_points: list[InterpolatedPair] = None
    _max_x: float = None
    _min_x: float = None
    _coordinates = None

    @property
    def max_x(self):
        if not self._max_x:
            self._max_x = max(self.coordinates, key = lambda coord: coord.x).x

        return self._max_x

    @property
    def min_x(self):
        if not self._min_x:
            self._min_x = min(self.coordinates, key = lambda coord: coord.x).x

        return self._min_x

    @property
    def coordinates(self):
        if not self._coordinates:
            coordinates = [pair_of_points.coordinate_a for pair_of_points in self.pairs_of_points]
            coordinates.append(self.pairs_of_points[-1].coordinate_b)

            self._coordinates = coordinates

        return self._coordinates

    def __init__(self, pairs_of_points: list[InterpolatedPair]):
        if not PythonValidator.is_list(pairs_of_points):
            if not PythonValidator.is_instance(pairs_of_points, InterpolatedPair):
                raise Exception('The provided "pairs_of_points" parameter is not a list of InterpolatedPair nor a single InterpolatedPair.')
            else:
                pairs_of_points = [pairs_of_points]

        for i in range(len(pairs_of_points) - 1):
            if pairs_of_points[i].coordinate_b.x != pairs_of_points[i + 1].coordinate_a.x or pairs_of_points[i].coordinate_b.y != pairs_of_points[i + 1].coordinate_a.y:
                raise Exception('The points must be consecutive, so the coordinate B of one point must be the coordinate A of the next one.')

        self.pairs_of_points = pairs_of_points

    def get_coord_from_x(self, x: float, is_normalized: bool = False, do_denormalize: bool = True):
        """
        Obtain the coordinate corresponding to the provided
        'x' in the whole graphic interpolation. This means 
        that the value will be interpolated with the pair of
        points that contains the provided 'x' in between.

        Use the 'do_denormalize' parameter to receive the
        coordinate denormalized or not.
        """
        # We are working with normalized coordinates, so if
        # the provided 'x' is not normalized, we need to do
        # it to be able to calculate
        if not is_normalized:
            # TODO: Really? Do we need the 'is_normalized' (?)
            x = Math.normalize(x, -10000, 10000)

        if not NumberValidator.is_number_between(x, self.min_x, self.max_x):
            raise Exception(f'The provided "x" parameter "{str(x)}" must be between {str(self.min_x)} and {str(self.max_x)}, both inclusive.')
        
        # We obtain the pair of points the 'x' belongs to
        # (it is in between of those 2 points)
        pair_of_points = next(pair_of_points for pair_of_points in self.pairs_of_points if pair_of_points.min_x <= x <= pair_of_points.max_x)

        coordinate = pair_of_points.get_coord_from_x(x)

        if do_denormalize:
            coordinate = Coordinate.denormalize_tuple((coordinate[0], coordinate[1]))

        return coordinate

    def get_coord_from_d(self, d: float, do_denormalize: bool = True):
        """
        Obtain the coordinate corresponding to the provided 
        distance 'd' in the whole graphic interpolation. This
        means that, if 5 pairs of points and 'd' is 0.3, it
        will return a coordinate located within the
        interpolation between the second pair of points.

        Consider 0.0 the start of the interpolation and 1.0 the
        end, and as we have 5 pair of points, the intervals 
        will be from [0, 0.2], [0.2, 0.4], [0.4, 0.6], 
        [0.6, 0.8], [0.8, 1.0]. As you can see, if 'd' is 0.3,
        the interval will be the second one.

        Use the 'do_denormalize' parameter to receive the
        coordinate denormalized or not.
        """
        if not NumberValidator.is_number_between(d, 0.0, 1.0):
            raise Exception(f'The provided "d" parameter "{str(d)}" must be between 0.0 and 1.0, both inclusive.')

        # 'd' is the distance within the whole graphic
        num_of_splines = len(self.pairs_of_points)
        spline_size = 1
        if num_of_splines > 1:
            spline_size = 1 / num_of_splines

        if d == 1.0:
            spline_index = num_of_splines - 1
        else:
            # Check which spline we need to use
            spline_index = int(d // spline_size)

        # If we receive 0.28 => 0.28 - 0.25 = 0.03 / spline_size
        # A distance of 0.03 in the global graph is lower than the
        # real distance in any splice
        # 0.28 = 0.28 % (1 * 0.25) / 0.25 = 0.03 / 0.25 = 0.12
        # so the 3% of the global graphic distance is actually a
        # 12% in spline distance because there are 4 splines
        if spline_index > 0:
            d = d % (spline_index * spline_size) / spline_size
        else:
            d = d / spline_size

        # spline_index is the coord, d is the percentage of
        # distance in the interpolation of the pair of coords
        pair_of_points = self.pairs_of_points[spline_index]

        # Obtain the corresponding x based on 'd' and points
        x = pair_of_points.coordinate_a.x + d * (pair_of_points.coordinate_b.x - pair_of_points.coordinate_a.x)

        coordinate = pair_of_points.get_coord_from_x(x)

        if do_denormalize:
            coordinate = Coordinate.denormalize_tuple((coordinate[0], coordinate[1]))

        return coordinate
    
    def plot(self, do_denormalize_values: bool = False):
        """
        This method is just for testing and shows the graphic
        using the matplotlib library.

        If you are plotting a path movement along the screen,
        set the 'do_denormalize_values' as True to make the
        plot show the real movement the video will do through
        the scene. If you are plotting a graph used for any
        other type of effect, keep the value as False.
        """
        # We draw 100 points per interpolation
        # TODO: Can I cast it as a tuple (?)
        graph = np.array([self.get_coord_from_d(d, False) for d in np.linspace(0, 1, len(self.coordinates) * 100)])
        points = np.array([coordinate.as_tuple() for coordinate in self.coordinates])

        if do_denormalize_values:
            # I denormalize values to be able to watch the path
            # movement properly
            # TODO: Import these min and max from constants
            graph = np.array([Coordinate.denormalize_tuple((graphi[0], graphi[1])) for graphi in graph])
            points = np.array([Coordinate.denormalize_tuple((pointi[0], pointi[1])) for pointi in points])

        plt.plot(graph[:, 0], graph[:, 1], label = 'Graphic representation', color = 'blue')
        plt.scatter(points[:, 0], points[:, 1], color = 'red', label = 'Control points')
        plt.legend()
        plt.title('Graphic representation')
        plt.xlabel('x')
        plt.ylabel('y')

        if do_denormalize_values:
            # I invert the axis to see the graph as the video
            # movement because denormalization if because of
            # this
            plt.gca().invert_yaxis()
            
        plt.grid(True)
        plt.show()


# TODO: Check moviepy.interpolation methods


"""
    # TODO: I think we should move this file to 
    'yta_multimedia.video.frames' instead.
"""
from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.consts import MOVIEPY_SCENE_DEFAULT_SIZE
from yta_multimedia.video.frames.video_frame import VideoFrame
from yta_general_utils.color import Color
from yta_general_utils.programming.parameter_validator import NumberValidator, PythonValidator
from moviepy import ColorClip, VideoClip
from typing import Union


class ClipGenerator:
    # TODO: Validate that this below is working because I've
    # never before used a transparent black background
    @staticmethod
    def get_default_background_video(size: tuple = MOVIEPY_SCENE_DEFAULT_SIZE, duration: float = 1 / 60, is_transparent: bool = True):
        """
        Default full black background video that lasts
        'duration' seconds that represents our default moviepy
        scenario of 1920x1080 dimensions. This background is
        used for the basic position calculations. Check the
        transparent option if you are planning to stack some
        videos.

        The background can be build with a full transparent
        mask if 'is_transparent' provided as True.
        """
        opacity = 0.0 if is_transparent else 1.0

        return ClipGenerator.generate_color_background(size, [0, 0, 0], duration, 60, opacity)
    
    @staticmethod
    def generate_color_background(size: tuple, color: Union[Color, 'ColorString', str], duration: float = 1 / 60, fps: float = 60.0, opacity: float = 1.0):
        color = Color.parse(color)
        ClipValidator.validate_size(size)
        ClipValidator.validate_duration(duration)
        ClipValidator.validate_fps(fps)
        ClipValidator.validate_opacity(opacity)
        
        color_clip: ColorClip = ColorClip(size, color.as_rgb_array(), duration = duration).with_fps(fps)

        # A full opaque clip doesn't need a mask because it is,
        # by definition, full opaque
        if opacity < 1.0:
            color_clip = color_clip.with_mask(ClipMaskGenerator.get_uniform_mask(color_clip.size, color_clip.duration, color_clip.fps, opacity))

        return color_clip

# TODO: Move to another file, please, but avoid 
# cyclic import issues
class ClipValidator:
    """
    Class to simplify parameter validation while working
    with moviepy clips and clip parameters.
    """
    @staticmethod
    def validate_size(size: tuple):
        if not PythonValidator.is_tuple(size) or len(size) != 2 or not NumberValidator.is_number_between(size[0], 0, 5000) or not NumberValidator.is_number_between(size[1], 0, 5000):
            raise Exception('The provided "duration" is not a tuple or does not have the expected size (2) or the values are not between 0 and 5000.')
        
        return True
    
    @staticmethod
    def validate_duration(duration: float):
        # TODO: We could maybe accept 'fps' parameter to check
        # if 'duration' is a multiple
        if not NumberValidator.is_positive_number(duration, False):
            raise Exception('The provided "duration" is not a positive number.')
        
        return True
    
    @staticmethod
    def validate_fps(fps: int):
        if not NumberValidator.is_positive_number(fps):
            raise Exception('The provided "fps" is not a positive number.')
        
        return True
    
    @staticmethod
    def validate_opacity(opacity: float):
        if not NumberValidator.is_number_between(opacity, 0.0, 1.0):
            raise Exception('The provided "opacity" is not a float value between 1.0 and 0.0.')
        
        return True

# TODO: Move to another file, please, but avoid 
# cyclic import issues
class ClipMaskGenerator:
    @staticmethod
    def get_uniform_mask(size: tuple = MOVIEPY_SCENE_DEFAULT_SIZE, duration: float = 1 / 60, fps: int = 60, opacity: float = 1.0):
        """
        Get a ColorClip of the 'opacity' provided, where 1.0 indicates
        full opaque and 0.0 full transparent, to be used as a mask. 
        
        This is called uniform because it uses a ColorClip that doesn't
        change along its duration time.
        """
        ClipValidator.validate_size(size)
        ClipValidator.validate_duration(duration)
        ClipValidator.validate_fps(fps)
        ClipValidator.validate_opacity(opacity)
        
        return ColorClip(size, opacity, is_mask = True, duration = duration).with_fps(fps)

    @staticmethod
    def get_uniform_opaque_mask(size: tuple = MOVIEPY_SCENE_DEFAULT_SIZE, duration: float = 1 / 60, fps: int = 60):
        """
        Get a mask clip of the provided 'size' that is completely
        opaque and can be set as the mask of any other normal clip.
        
        This is called uniform because it uses a ColorClip that 
        doesn't change along its duration time.
        """
        return ClipMaskGenerator.get_uniform_mask(size, duration, fps, 1)
    
    @staticmethod
    def get_uniform_transparent_mask(size: tuple = MOVIEPY_SCENE_DEFAULT_SIZE, duration: float = 1 / 60, fps: int = 60):
        """
        Get a maskclip  of the provided 'size' that is completely
        transparent and can be set as the mask of any other normal
        clip.

        This is called uniform because it uses a ColorClip that 
        doesn't change along its duration time.
        """
        return ClipMaskGenerator.get_uniform_mask(size, duration, fps, 0)
    
    @staticmethod
    def video_to_mask(video: VideoClip):
        """
        Turn the provided 'video' into a mask VideoClip that can be
        set as the mask of any other normal clip.
        """
        video = VideoParser.to_moviepy(video)

        # TODO: This is ok but very slow I think...
        mask_clip_frames = [VideoFrame(frame).as_mask() for frame in video.iter_frames()]

        return VideoClip(lambda t: mask_clip_frames[int(t * video.fps)], is_mask = True).with_fps(video.fps)




from yta_multimedia.video.edition.effect.m_effect import MEffect as Effect
from yta_multimedia.video.edition.effect.moviepy.objects import MoviepyWithPrecalculated
from yta_multimedia.video.edition.effect.moviepy.t_function import TFunctionResize
from yta_multimedia.video.consts import MOVIEPY_SCENE_DEFAULT_SIZE
from yta_multimedia.video import MPVideo
from yta_multimedia.video.position import Position
from yta_multimedia.video.edition.effect.moviepy.position.objects.coordinate import Coordinate
from yta_multimedia.video.edition.effect.moviepy.position.utils.factor import get_factor_to_fit_scene, get_factor_to_fit_area
from yta_multimedia.video.parser import VideoParser
from yta_general_utils.math.rate_functions import RateFunction
from yta_general_utils.programming.parameter_validator import PythonValidator
from moviepy.Clip import Clip
from typing import Union


class ZoomOutOnVideoPositionEffect(Effect):
    """
    Creates a Zoom using a specific position of the video
    as its center to make the zoom effect on it. This 
    effect will adapt the video size to fit the full scene
    size and avoid showing any black area.
    """
    def apply(cls, video: Clip, video_position: Union[tuple, Coordinate, Position], zoom_start_region_size: tuple, rate_func: type = RateFunction.linear):
        """
        Apply the effect on the provided 'video'. The 'video_position'
        parameter is the position (within the video) we want to use as
        the new center. This effect will place the center of the
        provided 'video' in the Position.CENTER position (center of 
        the scene).
        """
        # TODO: What about 'rate_func' (?)
        video = VideoParser.to_moviepy(video)

        if not PythonValidator.is_instance(video_position, [Position, Coordinate]):
            if not PythonValidator.is_instance(video_position, tuple) and len(video_position) != 2:
                raise Exception('Provided "video_position" is not a valid Position enum or (x, y) tuple.')
            else:
                video_position = Coordinate(video_position[0], video_position[1])

        if not PythonValidator.is_tuple(zoom_start_region_size) or len(zoom_start_region_size) != 2:
            raise Exception('The provided "zoom_start_region_size" parameter is not a valid tuple of 2 values.')
            
        # TODO: This is actually written in another part of the code
        # and maybe I can make it dynamic depending on the actual
        # background (scene) size, but by now here it is to make it 
        # work
        MAX_SCENE_SIZE = MOVIEPY_SCENE_DEFAULT_SIZE
        # We force Position.CENTER because any other thing doesn't
        # make sense at all (at least by now), but I can set it as
        # a parameter
        POSITION_IN_SCENE = Position.CENTER

        resize_factor = get_factor_to_fit_area(zoom_start_region_size, MAX_SCENE_SIZE)

        video_handler = MPVideo(video)
        resizes = [
            TFunctionResize.resize_from_to(t, video.duration, resize_factor, 1, RateFunction.linear)
            for t in video_handler.frames_time_moments
        ]

        # We usually use the center of the video to manipulate it
        # and to place that center in the position in which we 
        # want to place it, because that is how it should be done
        # to be able to place it later
        video_center_difference = (
            video.w / 2 - video_position.x,
            video.h / 2 - video_position.y
        )

        positions = [
            POSITION_IN_SCENE.get_moviepy_upper_left_corner_tuple((
                # TODO: Maybe I need to apply a rate_func here to make
                # it non-linear
                (video.w - video_center_difference[0] * 2 * (1 - i / len(resizes))) * resizes[i],
                (video.h - video_center_difference[1] * 2 * (1 - i / len(resizes))) * resizes[i]
            )) for i in range(len(resizes))
        ]

        return MoviepyWithPrecalculated().apply(video, with_position_list = positions, resized_list = resizes)

    def apply_over_video():
        pass
    # TODO: 'apply_over_video' here (?)

    # TODO: I want to keep this code below because it is another
    # way of building the video animation. This code below makes
    # a zoom out using not the whole video but only the size
    # that makes, with the coordinate still on the center of the 
    # scene, fit the maximum video size without showing black
    # borders
    def test(cls, video: Clip, video_position: Union[tuple, Coordinate, Position], position_in_scene: Union[tuple, Coordinate, Position], rate_func: type = RateFunction.linear):
        """
        Apply the effect on the provided 'video'. The 'video_position'
        parameter is the position (within the video) we want to use as
        the new center. The 'position_in_scene' is the position (within
        the scene) in which we want to place the new center of the
        provided 'video'.
        """
        # TODO: What about 'rate_func' (?)
        video = VideoParser.to_moviepy(video)

        if not PythonValidator.is_instance(video_position, [Position, Coordinate]):
            if not PythonValidator.is_instance(video_position, tuple) and len(video_position) != 2:
                raise Exception('Provided "video_position" is not a valid Position enum or (x, y) tuple.')
            else:
                video_position = Coordinate(video_position[0], video_position[1])
            
        if not PythonValidator.is_instance(position_in_scene, [Position, Coordinate]):
            if not PythonValidator.is_instance(position_in_scene, tuple) and len(position_in_scene) != 2:
                raise Exception('Provided "position_in_scene" is not a valid Position enum or (x, y) tuple.')
            else:
                position_in_scene = Coordinate(position_in_scene[0], position_in_scene[1])

        # TODO: This is actually written in another part of the code
        # and maybe I can make it dynamic depending on the actual
        # background (scene) size, but by now here it is to make it 
        # work
        MAX_SCENE_SIZE = MOVIEPY_SCENE_DEFAULT_SIZE

        resize_factor = get_factor_to_fit_scene(video_position, position_in_scene, (video.w, video.h), MAX_SCENE_SIZE)

        video_handler = MPVideo(video)
        #gsvideo = gsvideo.resized(min_resize_factor)
        # TODO: This 1.5 is the static zoom factor we are using
        # now but it is only for this zoom. Make it dynamic later
        resizes = [
            TFunctionResize.resize_from_to(t, video.duration, resize_factor * 1.5, resize_factor, RateFunction.linear)
            for t in video_handler.frames_time_moments
        ]

        # We usually use the center of the video to manipulate it
        # and to place that center in the position in which we 
        # want to place it, because that is how it should be done
        # to be able to place it later
        video_center_difference = (
            video.w / 2 - video_position.x,
            video.h / 2 - video_position.y
        )

        positions = [
            position_in_scene.get_moviepy_upper_left_corner_tuple((
                (video.w - video_center_difference[0] * 2) * resizes[i],
                (video.h - video_center_difference[1] * 2) * resizes[i]
            )) for i in range(len(resizes))
        ]

        return MoviepyWithPrecalculated().apply(video, with_position_list = positions, resized_list = resizes)


"""
TODO: Please, put some structure here and more comments
"""
from yta_multimedia.video.consts import MANIM_SCENE_DEFAULT_SIZE


MANIM_RESOURCES_FOLDER = 'resources/manim/'

# Manim core
MANDATORY_CONFIG_PARAMETER = 1
OPTIONAL_CONFIG_PARAMETER = 0
# I obtained manim dimensions from here: https://docs.manim.community/en/stable/faq/general.html#what-are-the-default-measurements-for-manim-s-scene
LEFT_MARGIN = -MANIM_SCENE_DEFAULT_SIZE[0] / 2
UP_MARGIN = MANIM_SCENE_DEFAULT_SIZE[1] / 2
STANDARD_HEIGHT = 1080
"""
This is our standard height (in pixels) as we will create animation
videos of 1920x1080.
"""
STANDARD_WIDTH = 1920
"""
This is our standard width (in pixels) as we will create animation
videos of 1920x1080.
"""




from yta_multimedia.video.position import Position
from random import randrange


def get_in_and_out_positions_as_list():
    """
    Returns a list of 2 elements containing the out edge from which
    the video will come into the screen, and the opposite edge to get
    out of the screen. This has been created to animate a random slide
    transition effect. The possibilities are horizontal, diagonal and
    vertical linear sliding transitions. The first element in the list
    is the initial position and the second one, the final position. 
    """
    rnd = randrange(0, 8)
    
    if rnd == 0:
        positions = [Position.OUT_RIGHT, Position.OUT_LEFT]
    elif rnd == 1:
        positions = [Position.OUT_TOP, Position.OUT_BOTTOM]
    elif rnd == 2:
        positions = [Position.OUT_BOTTOM, Position.OUT_TOP]
    elif rnd == 3:
        positions = [Position.OUT_TOP_LEFT, Position.OUT_BOTTOM_RIGHT] 
    elif rnd == 4:
        positions = [Position.OUT_TOP_RIGHT, Position.OUT_BOTTOM_LEFT]
    elif rnd == 5:
        positions = [Position.OUT_BOTTOM_LEFT, Position.OUT_TOP_RIGHT]
    elif rnd == 6:
        positions = [Position.OUT_BOTTOM_RIGHT, Position.OUT_TOP_LEFT]
    elif rnd == 7:
        positions = [Position.OUT_LEFT, Position.OUT_RIGHT]

    return positions


"""
Image edition module.

Interesting links below:
- https://www.geeksforgeeks.org/image-enhancement-techniques-using-opencv-python/
- https://www.geeksforgeeks.org/changing-the-contrast-and-brightness-of-an-image-using-python-opencv/
"""
from yta_general_utils.image.parser import ImageParser
from yta_general_utils.programming.parameter_validator import NumberValidator
from PIL import Image, ImageEnhance
from pillow_lut import load_cube_file
from typing import Union

import cv2
import numpy as np
import colorsys



# LUTS
# You create a new table in which you set new values
# for each of the 256 pixels, so the previous pixel
# color 0 won't be 0 again, it will be the new value
# set on its [0] position
from yta_general_utils.programming.enum import YTAEnum as Enum
class LutTable(Enum):
    """
    Image LUT tables definition to be able to handle them
    and apply to images to modify those images.
    """
    INVERSE = 'inverse'
    SQUARE_ROOT = 'square_root'
    CUBE = 'cube'

    def get_lut_table(self):
        """
        Obtain the LUT table array, which is a 2D table with 256
        indexes containing the pixel color in which the original
        color must be converted.
        """
        functions = {
            LutTable.INVERSE: lambda pixel: 255 - pixel,
            LutTable.SQUARE_ROOT: lambda pixel: (pixel * 255) ** (1 / 2),
            LutTable.CUBE: lambda pixel: (pixel ** 3) / (255 ** 2)
        }

        return np.array([functions[self](i) for i in range(256)], dtype = np.uint8)
        
    def apply_to_image(self, image: any) -> np.ndarray:
        """
        Apply the lut table to the provided image.

        Result is a numpy array in RGB format.
        """
        # We obtain a result in GBR format and we transform
        # it into an RGB
        return cv2.LUT(ImageParser.to_opencv(image), self.get_lut_table())[:, :, ::-1]
    
class ImageEditor:
    """
    Class to simplify and encapsulate all the functionality
    related to image edition.
    """
    
    @staticmethod
    def modify_color_temperature(image: Union[str, Image.Image, np.ndarray], factor: int = 0):
        return change_image_color_temperature(image, factor)

    @staticmethod
    def modify_color_hue(image: Union[str, Image.Image, np.ndarray], factor: int = 0):
        return change_image_color_hue(image, factor)
    
    @staticmethod
    def modify_brightness(image: Union[str, Image.Image, np.ndarray], factor: int = 0):
        return change_image_brightness(image, factor)

    @staticmethod
    def modify_contrast(image: Union[str, Image.Image, np.ndarray], factor: int = 0):
        return change_image_contrast(image, factor)
    
    @staticmethod
    def modify_sharpness(image: Union[str, Image.Image, np.ndarray], factor: int = 0):
        return change_image_sharpness(image, factor)
    
    @staticmethod
    def modify_white_balance(image: Union[str, Image.Image, np.ndarray], factor: int = 0):
        return change_image_white_balance(image, factor)
    
    @staticmethod
    def apply_lut(image: Union[str, Image.Image, np.ndarray], lut_table: LutTable):
        """
        Apply the 2D Lut table provided in the 'lut_table'
        parameter to the also given 'image'.

        Thanks to:
        - https://gist.github.com/blroot/b22abc23526af2711d92cc3b3f13b907
        """
        lut_table = LutTable.to_enum(lut_table)

        return lut_table.apply_to_image(image)
    
    @staticmethod
    def apply_3d_lut(image: Union[str, Image.Image, np.ndarray], lut_3d_filename: str):
        """
        Apply a 3D Lut table, which is loaded from the
        provided 'lut_3d_filename' .cube file, to the
        also given 'image'.

        Thanks to:
        - https://stackoverflow.com/questions/73341263/apply-3d-luts-cube-files-into-an-image-using-python
        """
        from yta_general_utils.file.checker import FileValidator

        if not FileValidator.is_file(lut_3d_filename):
            raise Exception('The "lut_3d_filename" provided is not a valid file.')
        # TODO: Improve the validation to check that is .cube
        
        return ImageParser.to_pillow(image).filter(load_cube_file(lut_3d_filename))



# TODO: This value must be a setting for the ImageEditor
# class that must be in another file so it can be obtain
# without any cyclic import issue related to ImageEditor
COLOR_TEMPERATURE_CHANGE_LIMIT = (-50, 50)

def change_image_color_temperature(image: any, factor: int = 0) -> np.ndarray:
    """
    Change the 'image' color temperature by the
    provided 'factor', that must be a value between
    [-50, 50].

    The color change consist of updating the red and
    blue values, where red is calid and blue is cold.
    Increasing the temperature means increasing the
    red color, and decreasing it, decreasing the blue
    color.
    """
    if not NumberValidator.is_number_between(factor, COLOR_TEMPERATURE_CHANGE_LIMIT[0], COLOR_TEMPERATURE_CHANGE_LIMIT[1]):
        raise Exception(f'The "factor" parameter provided is not a number between [{COLOR_TEMPERATURE_CHANGE_LIMIT[0]}, {COLOR_TEMPERATURE_CHANGE_LIMIT[1]}].')
    
    # The '.copy()' makes it writeable
    image = ImageParser.to_numpy(image).copy()

    if factor == 0:
        return image

    # We want the factor being actually a value between 0.50 and 1.50,
    # but multiplying by 1.5 is equal to divide by 0.75 so I need to
    # manually do this calculation to apply the formula correctly
    factor = 1 - (0.25 - normalize(factor, COLOR_TEMPERATURE_CHANGE_LIMIT[0], 0, 0, 0.25)) if factor < 0 else 1 + normalize(factor, 0, COLOR_TEMPERATURE_CHANGE_LIMIT[1], 0, 0.5)
    
    r, b = image[:, :, 0], image[:, :, 2]
    
    # Min and max values are 0 and 255
    r = np.clip(r * factor, 0, 255)
    b = np.clip(b / factor, 0, 255)
    
    # Reconstruimos la imagen con los canales modificados
    image[:, :, 0] = r
    image[:, :, 2] = b

    return image


# TODO: This value must be a setting for the ImageEditor
# class that must be in another file so it can be obtain
# without any cyclic import issue related to ImageEditor
COLOR_HUE_CHANGE_LIMIT = (-50, 50)
# These below are 2 functions to convert
# TODO: Please, move these functions to a method
# maybe in 'yta_general_utils' or in 'yta_multimedia'
rgb_to_hsv = np.vectorize(colorsys.rgb_to_hsv)
hsv_to_rgb = np.vectorize(colorsys.hsv_to_rgb)

def change_image_color_hue(image: any, factor: int = 0) -> np.ndarray:
    """
    Change the 'image' color hue by the provided
    'factor', that must be a value between [-50, 50].

    Colorize PIL image `original` with the given
    `factor` (hue within 0-360); returns another PIL image.
    """
    # TODO: This method is not working
    if not NumberValidator.is_number_between(factor, COLOR_HUE_CHANGE_LIMIT[0], COLOR_HUE_CHANGE_LIMIT[1]):
        raise Exception(f'The "factor" parameter provided is not a number between [{COLOR_HUE_CHANGE_LIMIT[0]}, {COLOR_HUE_CHANGE_LIMIT[1]}].')
    
    # The '.copy()' makes it writeable
    image = ImageParser.to_numpy(image).copy()

    factor = normalize(factor, COLOR_HUE_CHANGE_LIMIT[0], COLOR_HUE_CHANGE_LIMIT[1], 0, 360)
    
    # TODO: This code is not working well
    # TODO: This method is very very slow
    #arr = np.array(np.asarray(img).astype('float'))
    #r, g, b, a = np.rollaxis(image, axis = -1)
    print(image.size) # size is 6220800
    r, g, b = np.rollaxis(image, axis = -1)
    #r, g, b = np.moveaxis(image, -1, 0)
    h, s, v = rgb_to_hsv(r, g, b)
    h = factor / 360.0
    r, g, b = hsv_to_rgb(h, s, v)
    #arr = np.dstack((r, g, b, a))
    arr = np.dstack((r, g, b)).astype(np.uint8)
    print(arr.size) # size is 220800
    print(arr)

    # TODO: I don't like this line below
    return arr
    return Image.fromarray(arr.astype('uint8'), 'RGBA')

BRIGHTNESS_LIMIT = (-100, 100)

def change_image_brightness(image: any, factor: int = 0) -> np.ndarray:
    """
    Change the 'image' brightness by the provided
    'factor', that must be a value between [-100, 100].
    """
    image = ImageParser.to_pillow(image).copy()

    # TODO: Check factor is between the limit
    if not NumberValidator.is_number_between(factor, BRIGHTNESS_LIMIT[0], BRIGHTNESS_LIMIT[1]):
        raise Exception(f'The provided factor must be a number between [{BRIGHTNESS_LIMIT[0]}, {BRIGHTNESS_LIMIT[1]}].')

    # factor from -100 to 0 must be from 0.5 to 1
    # factor from 0 to 100 must be from 1 to 2
    factor = normalize(factor, BRIGHTNESS_LIMIT[0], 0, 0.5, 1.0) if factor <= 0 else normalize(factor, 0, BRIGHTNESS_LIMIT[1], 1.0, 2.0)

    image = ImageEnhance.Brightness(image).enhance(factor)

    return ImageParser.to_numpy(image)

CONTRAST_LIMIT = (-100, 100)

def change_image_contrast(image: any, factor: int = 0) -> np.ndarray:
    """
    Change the 'image' contrast by the provided
    'factor', that must be a value between [-100, 100].
    """
    image = ImageParser.to_pillow(image).copy()

    # TODO: Check factor is between the limit
    if not NumberValidator.is_number_between(factor, CONTRAST_LIMIT[0], CONTRAST_LIMIT[1]):
        raise Exception(f'The provided factor must be a number between [{CONTRAST_LIMIT[0]}, {CONTRAST_LIMIT[1]}].')

    # factor from -100 to 0 must be from 0.5 to 1
    # factor from 0 to 100 must be from 1 to 2
    factor = normalize(factor, CONTRAST_LIMIT[0], 0, 0.5, 1.0) if factor <= 0 else normalize(factor, 0, CONTRAST_LIMIT[1], 1.0, 2.0)

    image = ImageEnhance.Contrast(image).enhance(factor)

    return ImageParser.to_numpy(image)

SHARPNESS_LIMIT = (-100, 100)

def change_image_sharpness(image: any, factor: int = 0) -> np.ndarray:
    """
    Change the 'image' sharpness by the provided
    'factor', that must be a value between [-100, 100].

    A factor of -100 gives you a blurred image while
    a factor of 100 gives you a sharped image.
    """
    image = ImageParser.to_pillow(image).copy()

    # TODO: Check factor is between the limit
    if not NumberValidator.is_number_between(factor, SHARPNESS_LIMIT[0], SHARPNESS_LIMIT[1]):
        raise Exception(f'The provided factor must be a number between [{SHARPNESS_LIMIT[0]}, {SHARPNESS_LIMIT[1]}].')

    # factor from -100 to 0 must be from 0.5 to 1
    # factor from 0 to 100 must be from 1 to 2
    factor = normalize(factor, SHARPNESS_LIMIT[0], 0, 0.0, 1.0) if factor <= 0 else normalize(factor, 0, SHARPNESS_LIMIT[1], 1.0, 2.0)

    image = ImageEnhance.Sharpness(image).enhance(factor)

    return ImageParser.to_numpy(image)

WHITE_BALANCE_LIMIT = (-100, 100)

def change_image_white_balance(image: any, factor: int = 0) -> np.ndarray:
    """
    TODO: Explain
    
    The result is in RGB format.
    """
    # TODO: Apply factor -> 0.0 means no change
    image = ImageParser.to_opencv(image)
    image = cv2.cvtColor(image, cv2.COLOR_BGR2LAB)
    # Average of green channel
    avg_a = np.average(image[:, :, 1])
    # Average of red channel
    avg_b = np.average(image[:, :, 2])
    image[:, :, 1] = image[:, :, 1] - ((avg_a - 128) * (image[:, :, 0] / 255.0) * 1.2)
    image[:, :, 2] = image[:, :, 2] - ((avg_b - 128) * (image[:, :, 0] / 255.0) * 1.2)

    balanced_image = cv2.cvtColor(image, cv2.COLOR_LAB2RGB)

    return balanced_image


# TODO: Move this method to 'yta_general_utils' library
# and check how other normalize methods are working,
# maybe we can use this general method
def normalize(number: float, input_lower_limit: float, input_upper_limit: float, output_lower_limit: float = 0.0, output_upper_limit: float = 1.0):
    """
    Normalize the 'number' value to be between 'output_lower_limit'
    and 'output_upper_limit', according to the input provided, that
    is between the 'input_lower_limit' and 'input_upper_limit' 
    values.
    """
    if not NumberValidator.is_number(number) or not NumberValidator.is_number(input_lower_limit) or not NumberValidator.is_number(input_upper_limit) or not NumberValidator.is_number(output_lower_limit) or not NumberValidator.is_number(output_upper_limit):
        raise Exception('All the parameters must be numbers.')

    if not NumberValidator.is_number_between(number, input_lower_limit, input_upper_limit):
        raise Exception('The "number" parameter provided is not a number between the input limits provided.')
    
    if input_upper_limit <= input_lower_limit or output_upper_limit <= output_lower_limit:
        raise Exception('The upper limit must be greater than the lower limit.')
    
    return (number - input_lower_limit) / (input_upper_limit - input_lower_limit) * (output_upper_limit - output_lower_limit) + output_lower_limit


# from yta_general_utils.file.checker import file_is_audio_file
# from moviepy import AudioFileClip, CompositeAudioClip



# # TODO: Refactor this, please
# def append_audio_clip_to_audio_clip(base_audio_clip: AudioFileClip, new_audio_clip: AudioFileClip, start_time: float = 0.0, output_filename: str = None):
#     """
#     We receive a 'base_audio_clip' in which we will append the 'new_audio_clip'
#     starting at 'start_time' seconds since the first one started. A combined 
#     audio will be generated and returned.

#     If 'output_filename' is provided, it will be also written in that 
#     destination.
#     """
#     if start_time < 0:
#         start_time = 0

#     combined = CompositeAudioClip([base_audio_clip, new_audio_clip.with_start(start_time)])
#     # TODO: Why 441000 here below?
#     combined.fps = 44100

#     if output_filename:
#         combined.write_audiofile(output_filename)

#     return combined

# # TODO: Refactor this, please
# def append_audio_file_to_audio_file(base_audio_filename: str, new_audio_filename: str, start_time: float = 1.0, output_filename: str = None):
#     """
#     We receive an audio file called 'base_audio_filename' and we append the provided
#     'new_audio_filename' to that existing audio to sound over it. That second 
#     sound will start in the 'start_time' second of the first 'audio_filename'.

#     If 'output_filename' provided, it will write the new combined audio. If not,
#     it will only return the new CompositeAudioClip object.

#     # TODO: Is this method really necessary (?) I think it is not...
#     """
#     if not base_audio_filename:
#         return None
    
#     if not file_is_audio_file(base_audio_filename):
#         return None
    
#     if not new_audio_filename:
#         return None
    
#     if not file_is_audio_file(new_audio_filename):
#         return None
    
#     if start_time < 0:
#         return None

#     return append_audio_clip_to_audio_clip(AudioFileClip(base_audio_filename), AudioFileClip(new_audio_filename), start_time, output_filename)





from yta_multimedia.audio.utils.converter import validate_audio, audiosegment_to_audioclip, numpy_to_audiosegment, audioclip_to_audiosegment, numpy_to_audioclip
from yta_general_utils.temp import create_temp_filename
from yta_general_utils.file.filename import ensure_file_extension
from yta_general_utils.programming.enum import YTAEnum as Enum
from yta_general_utils.programming.parameter_validator import PythonValidator
from yta_general_utils.file.checker import FileValidator
from moviepy import AudioFileClip, AudioClip
from pydub import AudioSegment
from typing import Union

import numpy as np


class AudioExtension(Enum):
    """
    Enum class to encapsulate the accepted audio extensions for our
    system.
    """
    # TODO: Maybe interconnect with 'ffmpeg_handler.py' Enums
    MP3 = 'mp3'
    WAV = 'wav'
    M4A = 'm4a'
    WMA = 'wma'
    CD = 'cd'
    OGG = 'ogg'
    AIF = 'aif'
    # TODO: Check which extensions are valid for the AudioSegment
    # and the 'export' method to be able to classify AudioExtension
    # enums in AudioSegmentAudioExtension or similar because we
    # should also have AudioExtension for the FfmpegHandler...

class AudioConverter:
    """
    Class to simplify and encapsulate the functionality related to
    audio conversion.
    """
    @staticmethod
    def to(audio: Union[str, np.ndarray, AudioSegment, AudioClip], extension: AudioExtension, output_filename: str):
        """
        This method converts the provided 'audio' to a audio with
        the provided 'extension' by storing it locally as the 
        provided 'output_filename' (or as a temporary file if not
        provided), and returns the new audio and the filename.

        This method returns two values: audio, filename
        """
        audio = AudioConverter.to_audiosegment(audio)
        extension = AudioExtension.to_enum(audio)

        if not output_filename:
            # TODO: Replace this when not exporting needed
            output_filename = create_temp_filename(f'tmp_converted_sound.{extension.value}')
        else:
            output_filename = ensure_file_extension(output_filename, extension.value)

        audio.export(output_filename, format = extension.value)
        audio = AudioConverter.to_audiosegment(output_filename)

        return audio, output_filename

    @staticmethod
    def to_wav(audio: Union[str, np.ndarray, AudioSegment, AudioClip], output_filename: str):
        """
        This method converts the provided 'audio' to a wav audio
        by storing it locally as the provided 'output_filename'
        (or as a temporary file if not provided), and returns the
        new audio and the filename.

        This method returns two values: audio, filename
        """
        return AudioConverter.to(audio, AudioExtension.WAV, output_filename)
    
    @staticmethod
    def to_mp3(audio: Union[str, np.ndarray, AudioSegment, AudioClip], output_filename: str):
        """
        This method converts the provided 'audio' to a mp3 audio
        by storing it locally as the provided 'output_filename'
        (or as a temporary file if not provided), and returns the
        new audio and the filename.

        This method returns two values: audio, filename
        """
        return AudioConverter.to(audio, AudioExtension.MP3, output_filename)
    
    @staticmethod
    def to_audioclip(audio: Union[str, np.ndarray, AudioSegment, AudioClip], output_filename: Union[str, None] = None):
        validate_audio(audio)
        
        if PythonValidator.is_string(audio):
            if not FileValidator.file_is_audio_file(audio):
                raise Exception('Provided "audio" filename is not a valid audio file.')
            
            audio = AudioFileClip(audio)
        elif PythonValidator.is_instance(audio, np.ndarray):
            # TODO: Check this works
            # TODO: Create the util
            audio = numpy_to_audioclip(audio)
        elif PythonValidator.is_instance(audio, AudioSegment):
            audio = audiosegment_to_audioclip(audio)

        if output_filename is not None:
            # TODO: Validate 'output_filename'
            audio.write_audiofile(output_filename)

        return audio, output_filename
    
    @staticmethod
    def to_audiosegment(audio: Union[str, np.ndarray, AudioSegment, AudioClip], output_filename: Union[str, None] = None):
        """
        Forces the provided 'audio' to be a pydub AudioSegment
        and returns it if valid 'audio' provided or raises an
        Exception if not.
        """
        validate_audio(audio)

        if PythonValidator.is_string(audio):
            if not FileValidator.file_is_audio_file(audio):
                raise Exception('Provided "audio" filename is not a valid audio file.')
            
            audio = AudioSegment.from_file(audio)
        elif PythonValidator.is_instance(audio, np.ndarray):
            # TODO: Check this
            # TODO: What about sample_rate (?)
            audio = numpy_to_audiosegment(audio)
        elif PythonValidator.is_instance(audio, AudioClip):
            audio = audioclip_to_audiosegment(audio)
        # elif isinstance(audio, (AudioFileClip, CompositeAudioClip)):
        #     audio = cls.moviepy_audio_to_audiosegment(audio)

        if output_filename is not None:
            # TODO: Validate 'output_filename'
            # TODO: Use the extension, please
            audio.export(output_filename, format = 'wav')

        return audio, output_filename
    
    


from yta_multimedia.video.edition.effect.m_effect import MEffect as Effect
from yta_multimedia.video.position import Position
from yta_multimedia.video.edition.effect.moviepy.mask import ClipGenerator
from yta_multimedia.video.edition.effect.moviepy.position.objects.coordinate import Coordinate
from yta_multimedia.video.edition.effect.moviepy.objects import MoviepyWith, MoviepyArgument
from yta_multimedia.video.edition.effect.moviepy.t_function import TFunctionSetPosition
from yta_general_utils.math.rate_functions import RateFunction
from moviepy.Clip import Clip
from typing import Union


class ShakeAtPositionEffect(Effect):
    def apply(self, video: Clip, position: Union[Position, Coordinate, tuple] = Position.CENTER) -> Clip:
        # TODO: This is not working properly yet
        #PythonValidator.validate_method_params(BlurEffect.apply, locals(), ['video'])
        background_video = ClipGenerator.get_default_background_video()

        return self.apply_over_video(video, background_video, position)
    
    # TODO: What about this (?)
    def apply_over_video(self, video: Clip, background_video: Clip, position: Union[Position, Coordinate] = Position.CENTER) -> Clip:
        arg = MoviepyArgument(position, position, TFunctionSetPosition.linear_with_normal_shaking, RateFunction.linear)

        return MoviepyWith().apply_over_video(video, background_video, with_position = arg)
    

# TODO: I kee this code below just for a few commits to
# be able to think about its structure and check the
# difference with the one above
# class ShakeAtPosition(Effect):
#     """
#     This effect will blur the whole clip. The greater the
#     'blur_radius' is, the more blurred it becomes.
#     """
#     def apply(self, video: Clip, position: Union[Position, Coordinate, tuple] = Position.CENTER) -> Clip:
#         # TODO: This is not working properly yet
#         #PythonValidator.validate_method_params(BlurEffect.apply, locals(), ['video'])
#         background_video = ClipGenerator.get_default_background_video()

#         return self.apply_over_video(video, background_video, position)
    
#     # TODO: What about this (?)
#     def apply_over_video(self, video: Clip, background_video: Clip, position: Union[Position, Coordinate] = Position.CENTER) -> Clip:
#         video = VideoParser.to_moviepy(video)
#         background_video = VideoParser.to_moviepy(background_video)

#         # TODO: Validate position

#         if PythonValidator.is_tuple(position):
#             position = Coordinate(position[0], position[1])

#         # Turn position into upper left moviepy position
#         if PythonValidator.is_instance(position, Position):
#             position = position.get_moviepy_upper_left_corner_tuple(video.size, background_video.size)
#         elif PythonValidator.is_instance(position, Coordinate):
#             position = position.get_moviepy_upper_left_corner_tuple(video.size, background_video.size)

#         video_handler = MPVideo(video)
#         background_video = video_handler.prepare_background_clip(background_video)

#         # TODO: What if I want to shake it while moving (?)

#         video = video.with_position(lambda t: shake_movement(t, position[0], position[1])).with_start(0).with_duration(video.duration)
        
#         return CompositeVideoClip([
#             background_video,
#             video
#         ])


"""
Thanks for the inspiration:
https://www.geeksforgeeks.org/opencv-motion-blur-in-python/
"""
from yta_general_utils.image.parser import ImageParser
from yta_general_utils.image.converter import ImageConverter
from yta_general_utils.programming.enum import YTAEnum as Enum
from typing import Union

import cv2 
import numpy as np 


class MotionBlurDirection(Enum):
    """
    The direction we want to apply on the
    Motion Blur effect.
    """
    HORIZONTAL = 'horizontal'
    VERTICAL = 'vertical'
    DIAGONAL_TOP_RIGHT = 'diagonal_top_right'
    DIAGONAL_TOP_LEFT = 'diagonal_top_left'

def apply_motion_blur(image: any, kernel_size: int = 30, direction: MotionBlurDirection = MotionBlurDirection.HORIZONTAL, output_filename: Union[str, None] = None):
    """
    Apply a motion blur effect on the given 'image'
    using the provided 'kernel_size' (the greater
    this value is, the more motion blur effect we
    will get on the image). The motion blur can be
    applied in one 'direction', and the result image
    can be stored locally if 'output_filename' is
    provided.
    """
    direction = MotionBlurDirection.to_enum(direction)

    # TODO: 'image' param type must be refactored
    image = ImageConverter.numpy_image_to_opencv(ImageParser.to_numpy(image))
    
    kernel = np.zeros((kernel_size, kernel_size)) 

    if direction == MotionBlurDirection.HORIZONTAL:
        kernel[int((kernel_size - 1) / 2), :] = np.ones(kernel_size) 
    elif direction == MotionBlurDirection.VERTICAL:
        kernel[:, int((kernel_size - 1) / 2)] = np.ones(kernel_size) 
    elif direction == MotionBlurDirection.DIAGONAL_TOP_LEFT:
        np.fill_diagonal(kernel, 1)
    elif direction == MotionBlurDirection.DIAGONAL_TOP_RIGHT:
        np.fill_diagonal(np.fliplr(kernel), 1)
    
    # Normalize and apply
    kernel /= kernel_size
    image = cv2.filter2D(image, -1, kernel) 

    # TODO: Refactor this please
    if output_filename is not None:
        # Save output
        cv2.imwrite(output_filename, image)

    image = ImageConverter.opencv_image_to_numpy(image)

    return image

# TODO: Maybe this method has to be moved to 
# video handling
def get_progressive_kernel_sizes(steps: int, max: int):
    """
    Return an incremental array of 'steps' kernel sizes
    from 0 (excluded) to the provided 'max' value.
    
    This method is useful to apply an incremental effect.

    If steps = 20 and max = 100, it will return:
    [5, 10, 15, ..., 95, 100] (20 elements)
    """
    return np.linspace(0, max, steps + 1, endpoint = True, dtype = int)[1:]




"""
    There is some code here below that is generating a
    video moving throw the scene by the path defined in
    the GraphicInterpolation, and that is resized and
    the movement is recalculated based on the resized 
    so it is actually being resized while moving with
    no error.
"""
from yta_multimedia.video.edition.effect.moviepy.position.interpolation import GraphicInterpolation, InterpolationMethod, InterpolatedPair
from yta_multimedia.video.edition.effect.moviepy.objects import MoviepyWithPrecalculated
from yta_multimedia.video.position import Position
from yta_multimedia.video.edition.effect.moviepy.position.objects.coordinate import Coordinate
from moviepy import VideoFileClip


def test():
    # Create 3 consecutive pairs
    p1 = InterpolatedPair(Position.HALF_LEFT, Position.HALF_TOP, InterpolationMethod.LINEAR)
    p2 = InterpolatedPair(Position.HALF_TOP, Position.HALF_RIGHT, InterpolationMethod.LINEAR)
    p3 = InterpolatedPair(Position.HALF_RIGHT, Position.HALF_BOTTOM, InterpolationMethod.LINEAR)

    graph = GraphicInterpolation([p1, p2, p3])

    video = VideoFileClip('prueba.mp4').with_subclip(0, 1)
    #video = video.resized(0.5)
    fd = 1 / video.fps  # frame duration

    resizes = [1 - 0.7 * f * fd / video.duration for f in range(int(video.fps * video.duration))]
    # Obtain the position for each frame duration
    positions = [graph.get_coord_from_d(f * fd / video.duration) for f in range(int(video.fps * video.duration))]
    # The position above has been calculated for a video of
    # 1920x1080 over an scene of also 1920x1080, so adapt it to
    # our video

    # Transform center position to upper left corner
    for i, _ in enumerate(positions):
        positions[i] = Coordinate(positions[i][0], positions[i][1]).get_moviepy_upper_left_corner_tuple((video.w * resizes[i], video.h * resizes[i]))

    # TODO: I think we need to accept GraphicInterpolation as
    # a MoviepyWith parameter to be able to calculate the
    # positions for each frame and apply them considering the
    # also provided 'resizes'.
    # MoviepyWith is considering the 'resize' while moving, so
    # If i allow GraphicInterpolation to be accepted and adapt
    # the code to make the calculations, this example will be
    # working by itself just passing the GraphicInterpolation
    MoviepyWithPrecalculated.apply(video, with_position_list = positions, resized_list = resizes).write_videofile('a_test_precalculated.mp4')


"""
I'm placing the edition settings here until I find
a better place, but I want to keep them in the same
file by now to avoid duplicating them in the files
I am building for new concepts.

TODO: Move this to a definitive file in a near 
future, please
"""
# TODO: This must be read from another file which is
# not the 'image_editor' to avoid possible cyclic
# import issues
from yta_multimedia.image.edition.image_editor import COLOR_TEMPERATURE_CHANGE_LIMIT, BRIGHTNESS_LIMIT, CONTRAST_LIMIT, SHARPNESS_LIMIT, WHITE_BALANCE_LIMIT


MAX_TIMELINE_LAYER_DURATION = 1200
"""
The maximum duration a timeline layer can have
according to all the subclips on it. This value
can change to allow longer timeline layers.
"""

LAYERS_INDEXES_LIMIT = (0, 9)
"""
The limit of the layers indexes we have, starting
from 0, so only the upper limit + 1 layers are
available in the edition system.
"""
VOLUME_LIMIT = (0, 300)
"""
The limit of the volumen adjustment. Zero (0) means
silence, 100 means the original audio volume and
300 means 3 times higher volume.
"""
ZOOM_LIMIT = (1, 500)
"""
The limit of the zoom adjustment. One (1) means a
zoom out until the video is a 1% of its original 
size, 100 means the original size and 500 means a
zoom in to reach 5 times the original video size.
"""
COLOR_TEMPERATURE_LIMIT = COLOR_TEMPERATURE_CHANGE_LIMIT
"""
The limit of the color temperature adjustment. Zero
(0) means no change, while -50 means...

TODO: Fulfill this as it depends on another const
"""
BRIGHTNESS_LIMIT = BRIGHTNESS_LIMIT
"""
The limit of the image brightness adjustment.
"""
CONTRAST_LIMIT = CONTRAST_LIMIT
"""
The limit of the image contrast adjustment.
"""
SHARPNESS_LIMIT = SHARPNESS_LIMIT
"""
The limit of the image sharpness adjustment.
"""
WHITE_BALANCE_LIMIT = WHITE_BALANCE_LIMIT
"""
The limit of the image white balance adjustment.
"""


from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.frames.video_frame import VideoFrame
from yta_multimedia.video.edition.duration import set_video_duration
from yta_multimedia.video.edition.effect.moviepy.mask import ClipGenerator
from yta_multimedia.video.consts import MOVIEPY_SCENE_DEFAULT_SIZE
from yta_general_utils.programming.parameter_validator import PythonValidator, NumberValidator
from typing import Union
from moviepy import VideoClip, CompositeVideoClip, concatenate_videoclips
from moviepy.Clip import Clip
from copy import copy

import warnings
import random


SMALL_AMOUNT_TO_FIX = 0.0000000001
"""
Small amount we need to add to fix some floating
point number issues we've found. Something like
0.3333333333333326 will turn into 9 frames for a 
fps = 30 video, but this is wrong, as it should
be 10 frames and it is happening due to a minimal
floating point difference.
"""

class MPVideo:
    """
    Custom class to solve some bugs or issues with
    moviepy videos duration and frames number and
    malfunctioning related to floating points, and
    to also include some recurrent functionality.
    """
    _video: Clip = None
    """
    The original moviepy video clip.
    """
    _frames_indexes: list[int] = None
    _even_frames_indexes: list[int] = None
    _odd_frames_indexes: list[int] = None

    def __init__(self, video: Clip):
        self._video = video
        self.update_video_info()

    def update_video_info(self):
        """
        Process its own video to update the information
        if changed.
        """
        self._video = MPVideo.process_video(self._video)

    def __str__(self):
        return f'duration: {self.duration}\nfps: {self.fps}\nnumber_of_frames: {self.number_of_frames}\nframes_time_moments: {self.frames_time_moments}\nframes_indexes: {self.frames_indexes}\nframe_duration: {self.frame_duration}'

    @staticmethod
    def process_video(video: Clip):
        """
        Process the provided 'video' and makes it fit
        perfectly its real duration and forced to be
        60 fps.

        This method will try to do .get_frame() for 
        the last frames to obtain the real duration
        and adjust it consequently. It could make the
        execution slower, but we avoid issues with
        wrong durations.
        """
        video = VideoParser.to_moviepy(video)

        # 1. Ensure real duration and force to 60fps
        video = video.with_subclip(0, calculate_real_video_duration(video))#.with_fps(60)

        return video

    @property
    def video(self):
        """
        Original moviepy video clip.
        """
        return self._video

    @property
    def duration(self):
        """
        Real duration of the video, that has been
        checked according to the available frames.
        """
        return self.video.duration
    
    @property
    def fps(self):
        """
        Frames per second of the original video.
        """
        return self.video.fps
    
    @property
    def number_of_frames(self):
        """
        Number of frames of the moviepy video, using
        the previously checked duration.
        """
        return int(self.duration * self.fps)

    @property
    def frames_time_moments(self):
        """
        Array containing all the time moments of video
        frames, that can be used to obtain each frame
        individually with the 'video.get_frame(t = t)'
        method.

        Each frame time moment has been increased by 
        a small amount to ensure it is greater than 
        the base frame time value (due to decimals
        issue).
        """
        return get_frame_time_moments(self.duration, self.fps)
    
    @property
    def frames_indexes(self):
        """
        Array containing all the indexes of video frames,
        that can be used to obtain its corresponding
        frame time moment, or to make things simpler.
        """
        if self._frames_indexes is None:
            self._frames_indexes = get_frame_indexes(self.duration, self.fps)
        
        return self._frames_indexes
    
    @property
    def even_frames_indexes(self):
        """
        Array containing all the even indexes of video
        frames that can be used to obtain its corresponding
        frame time moment, or to simplify calculations.
        """
        if self._even_frames_indexes is None:
            self._even_frames_indexes = [
                frame_index
                for frame_index in self.frames_indexes
                if NumberValidator.is_even(frame_index)
            ]

        return self._even_frames_indexes

    @property
    def odd_frames_indexes(self):
        """
        Array containing all the odd indexes of video
        frames that can be used to obtain its corresponding
        frame time moment, or to simplify calculations.
        """
        if self._odd_frames_indexes is None:
            self._odd_frames_indexes = [
                frame_index
                for frame_index in self.frames_indexes
                if NumberValidator.is_odd(frame_index)
            ]

        return self._odd_frames_indexes
    
    @property
    def frame_duration(self):
        """
        Frame duration, based on video frames per second.
        """
        return 1 / self.fps
    
    # TODO: Maybe move this method to a VideoMaskInverter
    # class...
    def invert(self):
        """
        Invert the received 'video' (that must be a moviepy 
        mask or normal clip) and return it inverted as a
        VideoClip. If the provided 'video' is a mask, this 
        will be also a mask.

        If the 'clip' provided is a mask clip, remember to
        set it as the new mask of your main clip.

        This inversion is a process in which the numpy array
        values of each frame are inverted by substracting the
        highest value. If the frame is an RGB frame with 
        values between 0 and 255, it will be inverted by 
        doing 255 - X on each frame pixel value. If it is
        normalized and values are between 0 and 1 (it is a 
        mask clip frame), by doing 1 - X on each mask frame
        pixel value.
        """
        mask_frames = [VideoFrame(frame).inverted() for frame in self.video.iter_frames()]

        # TODO: Which calculation is better, t * fps or t // frame_duration (?)
        # TODO: What if we have fps like 29,97 (?) I proposed forcing
        # the videos to be 60fps always so we avoid this problem
        return VideoClip(lambda t: mask_frames[int(t * self.video.fps)], is_mask = self.video.is_mask).with_fps(self.video.fps)

    def prepare_background_clip(self, background_video: Union[str, Clip]):
        """
        Prepares the provided 'background_video' by modifying its duration to
        be the same as the provided 'video'. By default, the strategy is 
        looping the 'background_video' if the 'video' duration is longer, or
        cropping it if it is shorter. This method returns the background_clip
        modified according to the provided 'video'.

        This method will raise an Exception if the provided 'video' or the provided
        'background_video' are not valid videos.

        TODO: Add a parameter to be able to customize the extend or enshort
        background strategy.
        """
        background_video = VideoParser.to_moviepy(background_video)

        background_video = set_video_duration(background_video, self.video.duration)

        return background_video

    def wrap_with_transparent_background(self):
        return MPVideo.wrap_video_with_transparent_background(self.video)

    # TODO: Maybe do the same with frame times instead
    # of frame indexes
    def get_first_n_frames_indexes(self, n: int):
        """
        Obtain the first 'n' frames indexes of the current
        video to be able to use them within a condition
        with the '.get_frame(t)' method.

        This list can be used to check if the current frame
        number (index) is on it or not, to apply the frame
        image effect or to leave it as it is.

        Each frame time moment has been increased by 
        a small amount to ensure it is greater than 
        the base frame time value (due to decimals
        issue).
        """
        n = len(self.frames_indexes) if n > len(self.frames_indexes) else n

        return self.frames_indexes[:n]
    
    @staticmethod
    def get_last_n_frames_indexes(self, n: int):
        """
        Obtain the last 'n' frames indexes of the current
        video to be able to use them within a condition
        with the '.get_frame(t)' method.

        This list can be used to check if the current frame
        number (index) is on it or not, to apply the frame
        image effect or to leave it as it is.

        Each frame time moment has been increased by 
        a small amount to ensure it is greater than 
        the base frame time value (due to decimals
        issue).
        """
        n = len(self.frames_indexes) if n > len(self.frames_indexes) else n

        return self.frames_indexes[-n:]

    def get_first_n_even_frames_indexes(self, n: int):
        """
        Obtain the first 'n' even frames indexes of the 
        current video to be able to use them within a
        condition with the '.get_frame(t)' method.

        This list can be used to check if the current frame
        number (index) is on it or not, to apply the frame
        image effect or to leave it as it is.

        If 'n' is greater than the real number of even
        frames, 'n' will get that value so the result will
        be all the even frames indexes.

        Each frame time moment has been increased by 
        a small amount to ensure it is greater than 
        the base frame time value (due to decimals
        issue).
        """
        n = len(self.even_frames_indexes) if n > len(self.even_frames_indexes) else n

        return self.even_frames_indexes[:n]
    
    def get_last_n_even_frames_indexes(self, n: int):
        """
        Obtain the last 'n' even frames indexes of the
        current video to be able to use them within a
        condition with the '.get_frame(t)' method.

        This list can be used to check if the current frame
        number (index) is on it or not, to apply the frame
        image effect or to leave it as it is.

        If 'n' is greater than the real number of even
        frames, 'n' will get that value so the result will
        be all the even frames indexes.

        Each frame time moment has been increased by 
        a small amount to ensure it is greater than 
        the base frame time value (due to decimals
        issue).
        """
        n = len(self.even_frames_indexes) if n > len(self.even_frames_indexes) else n

        return self.even_frames_indexes[-n:]

    def get_first_n_odd_frames_indexes(self, n: int):
        """
        Obtain the first 'n' odd frames indexes of the 
        current video to be able to use them within a
        condition with the '.get_frame(t)' method.

        This list can be used to check if the current frame
        number (index) is on it or not, to apply the frame
        image effect or to leave it as it is.

        If 'n' is greater than the real number of odd
        frames, 'n' will get that value so the result will
        be all the odd frames indexes.

        Each frame time moment has been increased by 
        a small amount to ensure it is greater than 
        the base frame time value (due to decimals
        issue).
        """
        n = len(self.odd_frames_indexes) if n > len(self.odd_frames_indexes) else n

        return self.odd_frames_indexes[:n]
    
    def get_last_n_odd_frames_indexes(self, n: int):
        """
        Obtain the last 'n' odd frames indexes of the
        current video to be able to use them within a
        condition with the '.get_frame(t)' method.

        This list can be used to check if the current frame
        number (index) is on it or not, to apply the frame
        image effect or to leave it as it is.

        If 'n' is greater than the real number of odd
        frames, 'n' will get that value so the result will
        be all the odd frames indexes.

        Each frame time moment has been increased by 
        a small amount to ensure it is greater than 
        the base frame time value (due to decimals
        issue).
        """
        n = len(self.odd_frames_indexes) if n > len(self.odd_frames_indexes) else n

        return self.odd_frames_indexes[-n:]
    
    def get_n_random_frames_indexes(self, n: int):
        """
        Obtain a list of 'n' random frame indexes of the
        current video to be able to use them within a
        condition with the '.get_frame(t)' method.

        This list can be used to check if the current frame
        number (index) is on it or not, to apply the frame
        image effect or to leave it as it is.

        If 'n' is greater than the real number of frames,
        'n' will get that value so the result will be all
        the frames indexes but in a random order.

        Each frame time moment has been increased by 
        a small amount to ensure it is greater than 
        the base frame time value (due to decimals
        issue).
        """
        n = len(self.frames_indexes) if n > len(self.frames_indexes) else n
        
        return random.sample(self.frames_indexes, n)

    # TODO: I think static methods below must be
    # in another file because this class was born
    # to be used as instances
    @staticmethod
    def wrap_video_with_transparent_background(video: Clip):
        """
        Put a full transparent background behind the video
        if its size is not our default scene (1920x1080) and
        places the video on the center of this background.

        This method works with a copy of the original video so
        only the returned one is changed.
        """
        video = VideoParser.to_moviepy(video)

        video = copy(video)
        # TODO: Is this changing the variable value or do I
        # need to do by position (?)
        if video.size != MOVIEPY_SCENE_DEFAULT_SIZE:
            # I place the video at the center of the new background but
            # I reposition it to place with its center in the same place
            # as the original one
            original_center_positions = []
            # TODO: Careful with this t
            for t in get_frame_time_moments(video.duration, video.fps):
                pos = video.pos(t)
                original_center_positions.append((pos[0], pos[1]))

            video = CompositeVideoClip([
                ClipGenerator.get_default_background_video(duration = video.duration),
                video.with_position(('center', 'center'))
            ]).with_position(lambda t: (
                original_center_positions[MPVideo.frame_time_to_frame_index(t, 1 / video.fps)][0] - MOVIEPY_SCENE_DEFAULT_SIZE[0] / 2,
                original_center_positions[MPVideo.frame_time_to_frame_index(t, 1 / video.fps)][1] - MOVIEPY_SCENE_DEFAULT_SIZE[1] / 2
            ))

        return video

    @staticmethod
    def concatenate_videos(videos: list[Clip]):
        """
        Concatenate the provided 'videos' but fixing the
        videos dimensions. It will wrap any video that
        doesn't fit the 1920x1080 scene size with a full
        transparent background to fit those dimensions.
        """
        if not PythonValidator.is_list(videos):
            videos = [videos]

        videos = [
            MPVideo.wrap_with_transparent_background(video) 
            for video in [VideoParser.to_moviepy(video) for video in videos]
        ]
        
        return concatenate_videoclips(videos)

    @staticmethod
    def frame_time_to_frame_index(t: float, fps: float):
        """
        Transform the provided 't' frame time to 
        its corresponding frame index according
        to the 'fps' provided.

        This method applies the next formula:

        int(t * fps + SMALL_AMOUNT_TO_FIX)
        """
        return int((t + SMALL_AMOUNT_TO_FIX) * fps )

    @staticmethod
    def frame_index_to_frame_time(i: int, fps: float):
        """
        Transform the provided 'i' frame index to
        its corresponding frame time according to
        the 'fps' provided.

        This method applies the next formula:

        i * 1 / fps + SMALL_AMOUNT_TO_FIX
        """
        return i * 1 / fps + SMALL_AMOUNT_TO_FIX



def get_frame_indexes(duration, fps):
    """
    Get the list of the frame indexes of the video
    with the given 'duration' and 'fps'.

    If a video lasts 1 second and has fps = 5, this
    method will return: 0, 1, 2, 3, 4.
    """
    return [i for i in range(int(duration * fps + SMALL_AMOUNT_TO_FIX))]

def get_frame_time_moments(duration, fps):
    """
    Get the time moment of each video frame according
    to the provided video 'duration' and 'fps'. This
    will always include the second 0 and the 
    inmediately before the duration.

    If a video lasts 1 second and has fps = 5, this
    method will return: 0, 0.2, 0.4, 0.6, 0.8

    This method can return non-exact decimal values 
    so we recommend you to add a small amount to
    ensure it is above the expected base frame time.
    """
    return [(1 / fps * i) + SMALL_AMOUNT_TO_FIX for i in range(int(duration * fps + SMALL_AMOUNT_TO_FIX) + 1)][:-1]

def calculate_real_video_duration(video: Clip):
    """
    Process the provided 'video' and obtain the real
    duration by trying to access to the last frames
    according to its duration attribute.

    This method will return the real duration, which
    is determined by the last accessible frame plus
    the frame duration and a small amount to avoid
    decimal issues.
    """
    video = VideoParser.to_moviepy(video)

    # Moviepy library is throwing a warning when a 
    # frame is not accessible through its 
    # ffmpeg_reader, but will return the previous
    # valid frame and throw no Exceptions. As we
    # are trying to determine its real duration,
    # that warning is saying that the frame is not
    # valid, so it is not part of its real duration
    # so we must avoid it and continue with the
    # previous one until we find the first (last)
    # valid frame.
    warnings.filterwarnings('error')
    
    for t in get_frame_time_moments(video.duration, video.fps)[::-1]:
        try:
            # How to catch warnings: https://stackoverflow.com/a/30368735
            video.get_frame(t = t)
            last_frame = t
            break
        except:
            pass

    warnings.resetwarnings()

    frame_duration = 1 / video.fps
    # I sum a small amount to ensure it is over the
    # duration that guarantees the expected amount
    # of frames when calculating
    return ((last_frame + frame_duration + SMALL_AMOUNT_TO_FIX) // frame_duration) * frame_duration + SMALL_AMOUNT_TO_FIX



from yta_multimedia.video.generation.manim.classes.base_three_d_manim_animation import BaseThreeDManimAnimation
from yta_multimedia.video.generation.manim.classes.base_manim_animation_wrapper import BaseManimAnimationWrapper
from yta_multimedia.video.generation.manim.enums import MANIM_RENDERER
from manim import *


class Axes3DExampleWrapper(BaseManimAnimationWrapper):
    def __init__(self, text: str, duration: float):
        # TODO: This parameter validation could be done
        # by our specific parameter validator (under
        # construction yet)
        exception_messages = []

        if len(exception_messages) > 0:
            raise Exception('\n'.join([exception_message for exception_message in exception_messages]))
        
        super().__init__(Axes3DExampleGenerator)

class Axes3DExampleGenerator(BaseThreeDManimAnimation):
    def construct(self):
        """
        This method is called by manim when executed by shell and
        will call the scene animation render method to be processed
        and generated.
        """
        self.animate()

    def generate(self, parameters: dict, output_filename: Union[str, None] = None):
        """
        Build the manim animation video and stores it
        locally as the provided 'output_filename', 
        returning the final output filename.

        This method will call the '.animate()' method
        to build the scene with the instructions that
        are written in that method.
        """
        return super().generate(
            parameters,
            renderer = MANIM_RENDERER.CAIRO,
            output_filename = output_filename
        )
    
    def animate(self):
        axes = ThreeDAxes()

        x_label = axes.get_x_axis_label(Tex("x"))
        y_label = axes.get_y_axis_label(Tex("y")).shift(UP * 1.8)

        # 3D variant of the Dot() object
        dot = Dot3D()

        # zoom out so we see the axes
        self.set_camera_orientation(zoom=0.5)

        self.play(FadeIn(axes), FadeIn(dot), FadeIn(x_label), FadeIn(y_label))

        self.wait(0.5)

        # animate the move of the camera to properly see the axes
        self.move_camera(phi=75 * DEGREES, theta=30 * DEGREES, zoom=1, run_time=1.5)

        # built-in updater which begins camera rotation
        self.begin_ambient_camera_rotation(rate=0.15)

        # one dot for each direction
        upDot = dot.copy().set_color(RED)
        rightDot = dot.copy().set_color(BLUE)
        outDot = dot.copy().set_color(GREEN)

        self.wait(1)

        self.play(
            upDot.animate.shift(UP),
            rightDot.animate.shift(RIGHT),
            outDot.animate.shift(OUT),
        )

        self.wait(2)


class Image3DWrapper(BaseManimAnimationWrapper):
    def __init__(self, text: str, duration: float):
        # TODO: This parameter validation could be done
        # by our specific parameter validator (under
        # construction yet)
        exception_messages = []

        if len(exception_messages) > 0:
            raise Exception('\n'.join([exception_message for exception_message in exception_messages]))
        
        super().__init__(Image3DGenerator)

class Image3DGenerator(BaseThreeDManimAnimation):
    def construct(self):
        """
        This method is called by manim when executed by shell and
        will call the scene animation render method to be processed
        and generated.
        """
        self.animate()

    def generate(self, parameters: dict, output_filename: Union[str, None] = None):
        """
        Build the manim animation video and stores it
        locally as the provided 'output_filename', 
        returning the final output filename.

        This method will call the '.animate()' method
        to build the scene with the instructions that
        are written in that method.
        """
        return super().generate(
            parameters,
            renderer = MANIM_RENDERER.CAIRO,
            output_filename = output_filename
        )
    
    def animate(self):
        self.video1 = ImageMobject(
                filename_or_array = 'C:/Users/dania/Desktop/wallpaper1080.png',
            ).scale_to_fit_height(3)
        ax = Axes(
            x_range=[0, 10, 1],
            x_length=9,
            y_range=[0, 20, 5],
            y_length=6,
            axis_config={"include_numbers": True, "include_tip": False},

        ).to_edge(DL + RIGHT + UP, buff=1).scale(0.7)
        labels = ax.get_axis_labels()

        self.play(Create(VGroup(ax, labels)))
        self.play(FadeIn(self.video1))
        self.wait(3)
        self.move_camera(phi=0*DEGREES, theta= -90 * DEGREES, zoom= 0.7, run_time=0.4, gamma=0*DEGREES)

        self.begin_ambient_camera_rotation(90 * DEGREES / 3, about='phi')
        self.begin_ambient_camera_rotation(90 * DEGREES / 3, about='theta')
        self.wait(3)






# TODO: I don't know why these classes below only
# have a 'construct' method. I think it was to 
# simplify the way I rendered them for testing, but
# I think they should be like a 2D scene
# TODO: Remove these classes below soon, please
# class Axes3DExample(BaseThreeDManimAnimation):
#     def construct(self):
#         axes = ThreeDAxes()

#         x_label = axes.get_x_axis_label(Tex("x"))
#         y_label = axes.get_y_axis_label(Tex("y")).shift(UP * 1.8)

#         # 3D variant of the Dot() object
#         dot = Dot3D()

#         # zoom out so we see the axes
#         self.set_camera_orientation(zoom=0.5)

#         self.play(FadeIn(axes), FadeIn(dot), FadeIn(x_label), FadeIn(y_label))

#         self.wait(0.5)

#         # animate the move of the camera to properly see the axes
#         self.move_camera(phi=75 * DEGREES, theta=30 * DEGREES, zoom=1, run_time=1.5)

#         # built-in updater which begins camera rotation
#         self.begin_ambient_camera_rotation(rate=0.15)

#         # one dot for each direction
#         upDot = dot.copy().set_color(RED)
#         rightDot = dot.copy().set_color(BLUE)
#         outDot = dot.copy().set_color(GREEN)

#         self.wait(1)

#         self.play(
#             upDot.animate.shift(UP),
#             rightDot.animate.shift(RIGHT),
#             outDot.animate.shift(OUT),
#         )

#         self.wait(2)

# class Image3D(BaseThreeDManimAnimation):
#     def construct(self):
#         self.video1 = ImageMobject(
#                 filename_or_array = 'C:/Users/dania/Desktop/wallpaper1080.png',
#             ).scale_to_fit_height(3)
#         ax = Axes(
#             x_range=[0, 10, 1],
#             x_length=9,
#             y_range=[0, 20, 5],
#             y_length=6,
#             axis_config={"include_numbers": True, "include_tip": False},

#         ).to_edge(DL + RIGHT + UP, buff=1).scale(0.7)
#         labels = ax.get_axis_labels()

#         self.play(Create(VGroup(ax, labels)))
#         self.play(FadeIn(self.video1))
#         self.wait(3)
#         self.move_camera(phi=0*DEGREES, theta= -90 * DEGREES, zoom= 0.7, run_time=0.4, gamma=0*DEGREES)

#         self.begin_ambient_camera_rotation(90 * DEGREES / 3, about='phi')
#         self.begin_ambient_camera_rotation(90 * DEGREES / 3, about='theta')
#         self.wait(3)